import base64
import functools
import logging
import os
import shutil
import stat
import tempfile
from collections.abc import Collection, Sequence
from datetime import datetime
from itertools import chain
from mimetypes import guess_extension, guess_type
from pathlib import Path
from urllib.parse import quote as urlquote
from uuid import uuid4
from xml.etree import ElementTree as ET

import aiohttp
import thumbhash
from PIL import Image, ImageOps
from slixmpp import JID, Iq, Message
from slixmpp.plugins.xep_0264.stanza import Thumbnail
from slixmpp.plugins.xep_0447.stanza import StatelessFileSharing

from ...db.avatar import avatar_cache
from ...db.models import Attachment
from ...util.types import (
    LegacyAttachment,
    LegacyMessageType,
    LegacyThreadType,
    MessageReference,
)
from ...util.util import fix_suffix
from .. import config
from .message_text import TextMessageMixin


class AttachmentMixin(TextMessageMixin):
    PRIVILEGED_UPLOAD = False

    @property
    def __is_component(self) -> bool:
        return self.session is NotImplemented

    async def __upload(
        self,
        file_path: Path,
        file_name: str | None = None,
        content_type: str | None = None,
    ) -> str:
        assert config.UPLOAD_SERVICE

        file_size = file_path.stat().st_size

        content_type = content_type or self.xmpp.plugin["xep_0363"].default_content_type
        iq_slot = await self.__request_upload_slot(
            config.UPLOAD_SERVICE,
            file_name or file_path.name,
            file_size,
            content_type,
        )
        slot = iq_slot["http_upload_slot"]
        headers = {
            "Content-Length": str(file_size),
            "Content-Type": content_type,
            **{header["name"]: header["value"] for header in slot["put"]["headers"]},
        }

        async with aiohttp.ClientSession() as http:
            with file_path.open("rb") as fp:
                async with http.put(
                    slot["put"]["url"], data=fp, headers=headers
                ) as response:
                    response.raise_for_status()
        return slot["get"]["url"]

    async def __request_upload_slot(
        self,
        upload_service: JID | str,
        filename: str,
        size: int,
        content_type: str,
    ) -> Iq:
        iq_request = self.xmpp.make_iq_get(
            ito=upload_service, ifrom=config.UPLOAD_REQUESTER or self.xmpp.boundjid
        )
        request = iq_request["http_upload_request"]
        request["filename"] = filename
        request["size"] = str(size)
        request["content-type"] = content_type
        if self.__is_component or not self.PRIVILEGED_UPLOAD:
            return await iq_request.send()

        assert self.session is not NotImplemented
        iq_request.set_from(self.session.user_jid)
        return await self.xmpp["xep_0356"].send_privileged_iq(iq_request)

    @staticmethod
    async def __no_upload(
        file_path: Path,
        file_name: str | None = None,
        legacy_file_id: str | int | None = None,
    ) -> tuple[Path, str]:
        file_id = str(uuid4()) if legacy_file_id is None else str(legacy_file_id)
        assert config.NO_UPLOAD_PATH is not None
        assert config.NO_UPLOAD_URL_PREFIX is not None
        destination_dir = Path(config.NO_UPLOAD_PATH) / file_id

        if destination_dir.exists():
            log.debug("Dest dir exists: %s", destination_dir)
            files = list(f for f in destination_dir.glob("**/*") if f.is_file())
            if len(files) == 1:
                log.debug(
                    "Found the legacy attachment '%s' at '%s'",
                    legacy_file_id,
                    files[0],
                )
                name = files[0].name
                uu = files[0].parent.name  # anti-obvious url trick, see below
                return files[0], "/".join([file_id, uu, name])
            else:
                log.warning(
                    (
                        "There are several or zero files in %s, "
                        "slidge doesn't know which one to pick among %s. "
                        "Removing the dir."
                    ),
                    destination_dir,
                    files,
                )
                shutil.rmtree(destination_dir)

        log.debug("Did not find a file in: %s", destination_dir)
        # let's use a UUID to avoid URLs being too obvious
        uu = str(uuid4())
        destination_dir = destination_dir / uu
        destination_dir.mkdir(parents=True)

        name = file_name or file_path.name
        destination = destination_dir / name
        method = config.NO_UPLOAD_METHOD
        if method == "copy":
            shutil.copy2(file_path, destination)
        elif method == "hardlink":
            os.link(file_path, destination)
        elif method == "symlink":
            os.symlink(file_path, destination, target_is_directory=True)
        elif method == "move":
            shutil.move(file_path, destination)
        else:
            raise RuntimeError("No upload method not recognized", method)

        if config.NO_UPLOAD_FILE_READ_OTHERS:
            log.debug("Changing perms of %s", destination)
            destination.chmod(destination.stat().st_mode | stat.S_IROTH)
        uploaded_url = "/".join([file_id, uu, name])

        return destination, uploaded_url

    async def __valid_url(self, url: str) -> bool:
        async with self.session.http.head(url) as r:
            return r.status < 400

    async def __get_stored(self, attachment: LegacyAttachment) -> Attachment:
        if attachment.legacy_file_id is not None and not self.__is_component:
            with self.xmpp.store.session() as orm:
                stored = (
                    orm.query(Attachment)
                    .filter_by(
                        legacy_file_id=str(attachment.legacy_file_id),
                        user_account_id=self.session.user_pk,
                    )
                    .one_or_none()
                )
                if stored is not None:
                    if not await self.__valid_url(stored.url):
                        stored.url = None  # type:ignore
                    return stored
        return Attachment(
            user_account_id=None if self.__is_component else self.session.user_pk,
            legacy_file_id=None
            if attachment.legacy_file_id is None
            else str(attachment.legacy_file_id),
            url=attachment.url if config.USE_ATTACHMENT_ORIGINAL_URLS else None,
        )

    async def __get_url(
        self, attachment: LegacyAttachment, stored: Attachment
    ) -> tuple[bool, Path | None, str]:
        file_name = attachment.name
        content_type = attachment.content_type
        file_path = attachment.path

        if file_name and len(file_name) > config.ATTACHMENT_MAXIMUM_FILE_NAME_LENGTH:
            log.debug("Trimming long filename: %s", file_name)
            base, ext = os.path.splitext(file_name)
            file_name = (
                base[: config.ATTACHMENT_MAXIMUM_FILE_NAME_LENGTH - len(ext)] + ext
            )

        if file_path is None:
            if file_name is None:
                file_name = str(uuid4())
                if content_type is not None:
                    ext = guess_extension(content_type, strict=False)  # type:ignore
                    if ext is not None:
                        file_name += ext
            temp_dir = Path(tempfile.mkdtemp())
            file_path = temp_dir / file_name
            if attachment.url:
                async with self.session.http.get(attachment.url) as r:
                    r.raise_for_status()
                    with file_path.open("wb") as f:
                        f.write(await r.read())

            elif attachment.stream is not None:
                data = attachment.stream.read()
                if data is None:
                    raise RuntimeError

                with file_path.open("wb") as f:
                    f.write(data)
            elif attachment.aio_stream is not None:
                # TODO: patch slixmpp to allow this as data source for
                #       upload_file() so we don't even have to write anything
                #       to disk.
                with file_path.open("wb") as f:
                    async for chunk in attachment.aio_stream:
                        f.write(chunk)
            elif attachment.data is not None:
                with file_path.open("wb") as f:
                    f.write(attachment.data)

            is_temp = not bool(config.NO_UPLOAD_PATH)
        else:
            is_temp = False

        assert isinstance(file_path, Path)
        if config.FIX_FILENAME_SUFFIX_MIME_TYPE:
            file_name, content_type = fix_suffix(file_path, content_type, file_name)
            attachment.content_type = content_type
            attachment.name = file_name

        if config.NO_UPLOAD_PATH:
            local_path, new_url = await self.__no_upload(
                file_path, file_name, stored.legacy_file_id
            )
            new_url = (config.NO_UPLOAD_URL_PREFIX or "") + "/" + urlquote(new_url)
        else:
            local_path = file_path
            new_url = await self.__upload(file_path, file_name, content_type)
        if stored.legacy_file_id and new_url is not None:
            stored.url = new_url

        if local_path is not None and local_path.stat().st_size == 0:
            raise RuntimeError("File size is 0")

        return is_temp, local_path, new_url

    async def __set_sims(
        self,
        msg: Message,
        uploaded_url: str,
        path: Path | None,
        attachment: LegacyAttachment,
        stored: Attachment,
    ) -> Thumbnail | None:
        if stored.sims is not None:
            ref = self.xmpp["xep_0372"].stanza.Reference(xml=ET.fromstring(stored.sims))
            msg.append(ref)
            if ref["sims"]["file"].get_plugin("thumbnail", check=True):
                return ref["sims"]["file"]["thumbnail"]
            else:
                return None

        if not path:
            return None

        ref = self.xmpp["xep_0385"].get_sims(
            path, [uploaded_url], attachment.content_type, attachment.caption
        )
        if attachment.name:
            ref["sims"]["file"]["name"] = attachment.name
        thumbnail = None
        if attachment.content_type is not None and attachment.content_type.startswith(
            "image"
        ):
            try:
                h, x, y = await self.xmpp.loop.run_in_executor(
                    avatar_cache._thread_pool, get_thumbhash, path
                )
            except Exception as e:
                log.debug("Could not generate a thumbhash", exc_info=e)
            else:
                thumbnail = ref["sims"]["file"]["thumbnail"]
                thumbnail["width"] = x
                thumbnail["height"] = y
                thumbnail["media-type"] = "image/thumbhash"
                thumbnail["uri"] = "data:image/thumbhash;base64," + urlquote(h)

        stored.sims = str(ref)
        msg.append(ref)

        return thumbnail

    def __set_sfs(
        self,
        msg: Message,
        uploaded_url: str,
        path: Path | None,
        attachment: LegacyAttachment,
        stored: Attachment,
        thumbnail: Thumbnail | None = None,
    ) -> None:
        if stored.sfs is not None:
            msg.append(StatelessFileSharing(xml=ET.fromstring(stored.sfs)))
            return

        if not path:
            return

        sfs = self.xmpp["xep_0447"].get_sfs(
            path, [uploaded_url], attachment.content_type, attachment.caption
        )
        if attachment.name:
            sfs["file"]["name"] = attachment.name
        if attachment.disposition:
            sfs["disposition"] = attachment.disposition
        else:
            del sfs["disposition"]
        if thumbnail is not None:
            sfs["file"].append(thumbnail)
        stored.sfs = str(sfs)
        msg.append(sfs)

    async def __set_sfs_and_sims_without_download(
        self, msg: Message, attachment: LegacyAttachment
    ) -> None:
        assert attachment.url is not None

        if not any(
            (
                attachment.content_type,
                attachment.name,
                attachment.disposition,
            )
        ):
            return

        sims = self.xmpp.plugin["xep_0385"].stanza.Sims()
        ref = self.xmpp["xep_0372"].stanza.Reference()

        ref["uri"] = attachment.url
        ref["type"] = "data"
        sims["sources"].append(ref)
        sims.enable("file")

        xep_0447_stanza = self.xmpp.plugin["xep_0447"].stanza
        sfs = xep_0447_stanza.StatelessFileSharing()
        url_data = xep_0447_stanza.UrlData()
        url_data["target"] = attachment.url
        sfs["sources"].append(url_data)
        sfs.enable("file")

        if attachment.content_type:
            sims["file"]["media-type"] = attachment.content_type
            sfs["file"]["media-type"] = attachment.content_type
        if attachment.caption:
            sims["file"]["desc"] = attachment.caption
            sfs["file"]["desc"] = attachment.caption
        if attachment.name:
            sims["file"]["name"] = attachment.name
            sfs["file"]["name"] = attachment.name
        if attachment.disposition:
            sfs["disposition"] = attachment.disposition

        msg.append(sims)
        msg.append(sfs)

    def __send_url(
        self,
        msg: Message,
        legacy_msg_id: LegacyMessageType,
        uploaded_url: str,
        caption: str | None = None,
        carbon: bool = False,
        when: datetime | None = None,
        correction: bool = False,
        **kwargs,
    ) -> list[Message]:
        msg["oob"]["url"] = uploaded_url
        msg["body"] = uploaded_url
        if msg.get_plugin("sfs", check=True):
            msg["fallback"].enable("body")
            msg["fallback"]["for"] = self.xmpp.plugin["xep_0447"].stanza.NAMESPACE
        if caption:
            if correction:
                msg["replace"]["id"] = self._replace_id(legacy_msg_id)
            else:
                self._set_msg_id(msg, legacy_msg_id)
            m1 = self._send(msg, carbon=carbon, correction=correction, **kwargs)
            m2 = self.send_text(
                caption, legacy_msg_id=None, when=when, carbon=carbon, **kwargs
            )
            return [m1, m2] if m2 else [m1]
        else:
            if correction:
                msg["replace"]["id"] = self._replace_id(legacy_msg_id)
            else:
                self._set_msg_id(msg, legacy_msg_id)
            return [self._send(msg, carbon=carbon, **kwargs)]

    def __get_base_message(
        self,
        legacy_msg_id: LegacyMessageType | None = None,
        reply_to: MessageReference | None = None,
        when: datetime | None = None,
        thread: LegacyThreadType | None = None,
        carbon: bool = False,
        correction: bool = False,
        mto: JID | None = None,
    ) -> Message:
        if correction:
            xmpp_ids = self._legacy_to_xmpp(legacy_msg_id)
            if xmpp_ids:
                original_xmpp_id = xmpp_ids[0]
                for xmpp_id in xmpp_ids:
                    if xmpp_id == original_xmpp_id:
                        continue
                    self.retract(xmpp_id, thread)

        if reply_to is not None and reply_to.body:
            # We cannot have a "quote fallback" for attachments since most (all?)
            # XMPP clients will only treat a message as an attachment if the
            # body is the URL and nothing else.
            reply_to_for_attachment: MessageReference | None = MessageReference(
                reply_to.legacy_id, reply_to.author
            )
        else:
            reply_to_for_attachment = reply_to

        return self._make_message(
            when=when,
            reply_to=reply_to_for_attachment,
            carbon=carbon,
            mto=mto,
            thread=thread,
        )

    async def send_file(
        self,
        attachment: LegacyAttachment | Path | str,
        legacy_msg_id: LegacyMessageType | None = None,
        *,
        reply_to: MessageReference | None = None,
        when: datetime | None = None,
        thread: LegacyThreadType | None = None,
        **kwargs,
    ) -> tuple[str | None, list[Message]]:
        """
        Send a single file from this :term:`XMPP Entity`.

        :param attachment: The file to send.
            Ideally, a :class:`.LegacyAttachment` with a unique ``legacy_file_id``
            attribute set, to optimise potential future reuses.
            It can also be:
            - a :class:`pathlib.Path` instance to point to a local file, or
            - a ``str``, representing a fetchable HTTP URL.
        :param legacy_msg_id: If you want to be able to transport read markers from the gateway
            user to the legacy network, specify this
        :param reply_to: Quote another message (:xep:`0461`)
        :param when: when the file was sent, for a "delay" tag (:xep:`0203`)
        :param thread:
        """
        coro = self.__send_file(
            attachment,
            legacy_msg_id,
            reply_to=reply_to,
            when=when,
            thread=thread,
            **kwargs,
        )
        if self.__is_component:
            return await coro
        elif not isinstance(attachment, LegacyAttachment):
            return await coro
        elif attachment.legacy_file_id is None:
            return await coro
        else:
            # prevents race conditions where we download the same thing several time
            # and end up attempting to insert it twice in the DB, raising an
            # IntegrityError.
            async with self.session.lock(("attachment", attachment.legacy_file_id)):
                return await coro

    async def __send_file(
        self,
        attachment: LegacyAttachment | Path | str,
        legacy_msg_id: LegacyMessageType | None = None,
        *,
        reply_to: MessageReference | None = None,
        when: datetime | None = None,
        thread: LegacyThreadType | None = None,
        **kwargs,
    ) -> tuple[str | None, list[Message]]:
        store_multi = kwargs.pop("store_multi", True)
        carbon = kwargs.pop("carbon", False)
        mto = kwargs.pop("mto", None)
        correction = kwargs.get("correction", False)

        msg = self.__get_base_message(
            legacy_msg_id, reply_to, when, thread, carbon, correction, mto
        )

        if isinstance(attachment, str):
            attachment = LegacyAttachment(url=attachment)
        elif isinstance(attachment, Path):
            attachment = LegacyAttachment(path=attachment)

        stored = await self.__get_stored(attachment)

        if attachment.content_type is None and (
            name := (attachment.name or attachment.url or attachment.path)
        ):
            attachment.content_type, _ = guess_type(name)

        if stored.url:
            is_temp = False
            local_path = None
            new_url = stored.url
        else:
            try:
                is_temp, local_path, new_url = await self.__get_url(attachment, stored)
            except Exception as e:
                log.error("Error with attachment: %s: %s", attachment, e)
                log.debug("", exc_info=e)
                msg["body"] = (
                    f"/me tried to send a file ({attachment.format_for_user()}), "
                    f"but something went wrong: {e}. "
                )
                self._set_msg_id(msg, legacy_msg_id)
                return None, [self._send(msg, **kwargs)]
        assert new_url is not None

        stored.url = new_url
        if config.USE_ATTACHMENT_ORIGINAL_URLS and attachment.url:
            await self.__set_sfs_and_sims_without_download(msg, attachment)
        else:
            thumbnail = await self.__set_sims(
                msg, new_url, local_path, attachment, stored
            )
            self.__set_sfs(msg, new_url, local_path, attachment, stored, thumbnail)

        if not self.__is_component:
            with self.xmpp.store.session(expire_on_commit=False) as orm:
                orm.add(stored)
                orm.commit()

        if is_temp and isinstance(local_path, Path):
            local_path.unlink()
            local_path.parent.rmdir()

        msgs = self.__send_url(
            msg, legacy_msg_id, new_url, attachment.caption, carbon, when, **kwargs
        )
        if not self.__is_component:
            if store_multi:
                self.__store_multi(legacy_msg_id, msgs)
        return new_url, msgs

    def __send_body(
        self,
        body: str | None = None,
        legacy_msg_id: LegacyMessageType | None = None,
        reply_to: MessageReference | None = None,
        when: datetime | None = None,
        thread: LegacyThreadType | None = None,
        **kwargs,
    ) -> Message | None:
        if body:
            return self.send_text(
                body,
                legacy_msg_id,
                reply_to=reply_to,
                when=when,
                thread=thread,
                **kwargs,
            )
        else:
            return None

    async def send_files(
        self,
        attachments: Collection[LegacyAttachment],
        legacy_msg_id: LegacyMessageType | None = None,
        body: str | None = None,
        *,
        reply_to: MessageReference | None = None,
        when: datetime | None = None,
        thread: LegacyThreadType | None = None,
        body_first: bool = False,
        correction: bool = False,
        correction_event_id: LegacyMessageType | None = None,
        **kwargs,
    ) -> None:
        # TODO: once the epic XEP-0385 vs XEP-0447 battle is over, pick
        #       one and stop sending several attachments this way
        # we attach the legacy_message ID to the last message we send, because
        # we don't want several messages with the same ID (especially for MUC MAM)
        if not attachments and not body:
            # ignoring empty message
            return
        body_msg_id = (
            legacy_msg_id if body_needs_msg_id(attachments, body, body_first) else None
        )
        send_body = functools.partial(
            self.__send_body,
            body=body,
            reply_to=reply_to,
            when=when,
            thread=thread,
            correction=correction,
            legacy_msg_id=body_msg_id,
            correction_event_id=correction_event_id,
            **kwargs,
        )
        all_msgs = []
        if body_first:
            all_msgs.append(send_body())
        for i, attachment in enumerate(attachments):
            if i == 0 and body_msg_id is None:
                legacy = legacy_msg_id
            else:
                legacy = None
            _url, msgs = await self.send_file(
                attachment,
                legacy,
                reply_to=reply_to,
                when=when,
                thread=thread,
                store_multi=False,
                **kwargs,
            )
            all_msgs.extend(msgs)
        if not body_first:
            all_msgs.append(send_body())
        self.__store_multi(legacy_msg_id, all_msgs)

    def __store_multi(
        self,
        legacy_msg_id: LegacyMessageType | None,
        all_msgs: Sequence[Message | None],
    ) -> None:
        if legacy_msg_id is None:
            return
        ids = []
        for msg in all_msgs:
            if not msg:
                continue
            if stanza_id := msg.get_plugin("stanza_id", check=True):
                ids.append(stanza_id["id"])
            else:
                ids.append(msg.get_id())
        with self.xmpp.store.session() as orm:
            self.xmpp.store.id_map.set_msg(
                orm, self._recipient_pk(), str(legacy_msg_id), ids, self.is_participant
            )
            orm.commit()


def body_needs_msg_id(
    attachments: Collection[LegacyAttachment], body: str | None, body_first: bool
) -> bool:
    if attachments:
        return bool(body and body_first)
    else:
        return True


def get_thumbhash(path: Path) -> tuple[str, int, int]:
    with path.open("rb") as fp:
        img = Image.open(fp)
        width, height = img.size
        img = img.convert("RGBA")
        if width > 100 or height > 100:
            img.thumbnail((100, 100))
    img = ImageOps.exif_transpose(img)
    rgba_2d = list(img.getdata())
    rgba = list(chain(*rgba_2d))
    ints = thumbhash.rgba_to_thumb_hash(img.width, img.height, rgba)
    return base64.b64encode(bytes(ints)).decode(), width, height


log = logging.getLogger(__name__)



import logging
from pathlib import Path
from typing import Optional

import pytest
from slixmpp import JID

from slidge import main
from slidge.core import config
from slidge.util.conf import ConfigModule


def test_get_parser(monkeypatch):
    class Config:
        REQUIRED: str
        REQUIRED__DOC = "some doc"
        REQUIRED__SHORT = "r"

        REQUIRED_INT: int
        REQUIRED_INT__DOC = "some doc"

        MULTIPLE: tuple[str, ...] = ()
        MULTIPLE__DOC = "some more doc"

        OPTIONAL: Optional[str] = None
        OPTIONAL__DOC = "not required"

        SOME_BOOL = False
        SOME_BOOL__DOC = "a bool"

    monkeypatch.setattr(main, "config", Config)
    parser = main.get_parser()
    with pytest.raises(SystemExit) as e:
        parser.parse_known_args([])
    assert e.value.args[0] == 2  # Exit code 2

    args = parser.parse_args(["--required", "some_value", "--required-int", "45"])
    assert args.required == "some_value"
    assert args.required_int == 45

    args = parser.parse_args(["-r", "some_value", "--required-int", "45"])
    assert args.required == "some_value"
    assert args.required_int == 45
    assert args.multiple == tuple()

    args = parser.parse_args(
        ["-r", "some_value", "--required-int", "45", "--multiple", "a", "b"]
    )
    assert args.required == "some_value"
    assert args.required_int == 45
    assert args.multiple == ["a", "b"]
    assert args.optional is None

    args = parser.parse_args(
        [
            "-r",
            "some_value",
            "--required-int",
            "45",
            "--multiple",
            "a",
            "b",
            "--optional",
            "prout",
            "--some-bool",
        ]
    )
    assert args.required == "some_value"
    assert args.required_int == 45
    assert args.multiple == ["a", "b"]
    assert args.optional == "prout"
    assert args.some_bool


def test_bool(tmpdir, tmp_path):
    class Config:
        SOME_BOOL = False
        SOME_BOOL__DOC = "a bool"

        TRUE = True
        TRUE__DOC = "true by default"

    configurator = ConfigModule(Config)

    configurator.set_conf([])
    assert not Config.SOME_BOOL
    assert Config.TRUE

    configurator.set_conf(["--some-bool"])
    assert Config.SOME_BOOL
    assert Config.TRUE

    configurator.set_conf(["--true"])
    assert not Config.SOME_BOOL
    assert Config.TRUE

    configurator.set_conf(["--true=false"])
    assert not Config.SOME_BOOL
    assert not Config.TRUE

    configurator.set_conf(["--true=true"])
    assert not Config.SOME_BOOL
    assert Config.TRUE

    configurator.set_conf(["--some-bool=true"])
    assert Config.SOME_BOOL
    assert Config.TRUE

    configurator.set_conf(["--some-bool=false"])
    assert not Config.SOME_BOOL
    assert Config.TRUE

    # for the plugin-specific conf files, we use the rest
    configurator.parser.add_argument("-c", is_config_file=True)

    class Config2:
        SOME_OTHER_BOOL = False
        SOME_OTHER_BOOL__DOC = "a bool"

        TRUE2 = True
        TRUE2__DOC = "true by default"

    configurator2 = ConfigModule(Config2)
    conf_file = tmpdir / "conf.conf"

    # false
    conf_file.write_text("some-other-bool=false", "utf-8")
    args, rest = configurator.set_conf(["-c", str(conf_file)])
    assert rest
    configurator2.set_conf(rest)
    assert not Config2.SOME_OTHER_BOOL
    assert Config2.TRUE2

    # true
    conf_file.write_text("some-other-bool=true", "utf-8")
    args, rest = configurator.set_conf(["-c", str(conf_file)])
    assert rest
    configurator2.set_conf(rest)
    assert Config2.SOME_OTHER_BOOL
    assert Config2.TRUE2

    # true
    conf_file.write_text("", "utf-8")
    args, rest = configurator.set_conf(["-c", str(conf_file)])
    assert not rest
    configurator2.set_conf(rest)
    assert not Config2.SOME_OTHER_BOOL
    assert Config2.TRUE2

    conf_file.write_text("true2=true", "utf-8")
    args, rest = configurator.set_conf(["-c", str(conf_file)])
    assert rest
    configurator2.set_conf(rest)
    assert not Config2.SOME_OTHER_BOOL
    assert Config2.TRUE2

    conf_file.write_text("true2=false", "utf-8")
    args, rest = configurator.set_conf(["-c", str(conf_file)])
    assert rest
    configurator2.set_conf(rest)
    assert not Config2.SOME_OTHER_BOOL
    assert not Config2.TRUE2


def test_true_by_default_file(tmpdir, tmp_path):
    conf_file = tmpdir / "conf.conf"

    class Config:
        TRUE = True
        TRUE__DOC = "true by default"

    configurator = ConfigModule(Config)
    configurator.parser.add_argument("-c", is_config_file=True)

    conf_file.write_text("", "utf-8")
    args, rest = configurator.set_conf(["-c", str(conf_file)])
    assert not rest
    assert Config.TRUE

    # TODO: fix these cases!
    # conf_file.write_text("true=true", "utf-8")
    # args, rest = configurator.set_conf(["-c", str(conf_file)])
    # assert not rest
    # assert Config.TRUE

    # conf_file.write_text("true=false", "utf-8")
    # args, rest = configurator.set_conf(["-c", str(conf_file)])
    # assert not rest
    # assert not Config.TRUE


def test_false_by_default_file(tmpdir, tmp_path):
    conf_file = tmpdir / "conf.conf"

    class Config:
        FALSE = False
        FALSE__DOC = "true by default"

    configurator = ConfigModule(Config)
    configurator.parser.add_argument("-c", is_config_file=True)

    conf_file.write_text("", "utf-8")
    args, rest = configurator.set_conf(["-c", str(conf_file)])
    assert not rest
    assert not Config.FALSE

    conf_file.write_text("false=true", "utf-8")
    args, rest = configurator.set_conf(["-c", str(conf_file)])
    assert not rest
    assert Config.FALSE

    conf_file.write_text("false=false", "utf-8")
    args, rest = configurator.set_conf(["-c", str(conf_file)])
    assert not rest
    assert not Config.FALSE


def test_slidge_conf():
    args, rest = main.get_parser().parse_known_args(
        [
            "-c",
            str(Path(__file__).parent.parent / "dev" / "confs" / "slidge-example.ini"),
            "--legacy-module=slidge.plugins.dummy",
            "--jid=test.localhost",
            "--some-other",
        ]
    )
    assert args.server == "localhost"
    assert args.admins == ["test@localhost"]
    assert args.secret == "secret"
    assert args.loglevel == logging.DEBUG
    assert len(rest) == 1
    assert rest[0] == "--some-other"


def test_set_conf(monkeypatch):
    monkeypatch.setenv("SLIDGE_USER_JID_VALIDATOR", "cloup")
    argv = [
        "-c",
        str(Path(__file__).parent.parent / "dev" / "confs" / "slidge-example.ini"),
        "--legacy-module=slidge.plugins.dummy",
        "--jid=test.localhost",
        "--ignore-delay-threshold=200",
    ]
    main.get_configurator().set_conf(argv)
    assert config.SERVER == "localhost"
    assert config.ADMINS == ["test@localhost"]
    assert isinstance(config.ADMINS[0], JID)
    assert isinstance(config.JID, JID)
    assert config.SECRET == "secret"
    assert config.IGNORE_DELAY_THRESHOLD == 200
    assert config.USER_JID_VALIDATOR == "cloup"


def test_rest(tmp_path):
    class Config1:
        PROUT = "caca"
        PROUT__DOC = "?"

    class Config2:
        PROUT2: Optional[str] = None
        PROUT2__DOC = "?"

    configurator = ConfigModule(Config1)
    configurator.parser.add_argument("-c", is_config_file=True)
    conf_path = tmp_path / "test.conf"
    conf_path.write_text("prout2=something")
    args, rest = configurator.set_conf(["-c", str(conf_path)])

    configurator2 = ConfigModule(Config2)
    configurator2.set_conf(rest)

    assert Config1.PROUT == "caca"
    assert Config2.PROUT2 == "something"


def test_unrecognized_after_boolean():
    class Config:
        PROUT = True
        PROUT__DOC = "?"


    configurator = ConfigModule(Config)
    ns, rest = configurator.set_conf(["--prout", "--joujoujou"])
    assert Config.PROUT
    assert rest == ["--joujoujou"]


def test_custom_conf_dirs(monkeypatch, tmp_path_factory):
    dir1 = tmp_path_factory.mktemp("dir1")
    dir2 = tmp_path_factory.mktemp("dir2")

    (dir1 / "file1.ini").write_text("jid=some-jid\nlegacy-module=prout")
    (dir2 / "file2.ini").write_text("secret=some-secret")

    monkeypatch.setenv("SLIDGE_CONF_DIR", f"{dir1}:{dir2}")
    conf = main.get_configurator()
    args = conf.parser.parse_args([])
    assert args.jid == "some-jid"
    assert args.secret == "some-secret"
    assert args.legacy_module == "prout"



import logging
from copy import copy
from xml.etree import ElementTree

from slixmpp import JID, Message
from slixmpp.exceptions import XMPPError

from ....contact.contact import LegacyContact
from ....group.participant import LegacyParticipant
from ....group.room import LegacyMUC
from ....util.types import LinkPreview, Recipient
from ....util.util import dict_to_named_tuple, remove_emoji_variation_selector_16
from ... import config
from ...session import BaseSession
from ..util import DispatcherMixin, exceptions_to_xmpp_errors


class MessageContentMixin(DispatcherMixin):
    __slots__: list[str] = []

    def __init__(self, xmpp) -> None:
        super().__init__(xmpp)
        xmpp.add_event_handler("legacy_message", self.on_legacy_message)
        xmpp.add_event_handler("message_correction", self.on_message_correction)
        xmpp.add_event_handler("message_retract", self.on_message_retract)
        xmpp.add_event_handler("groupchat_message", self.on_groupchat_message)
        xmpp.add_event_handler("reactions", self.on_reactions)

    async def on_groupchat_message(self, msg: Message) -> None:
        await self.on_legacy_message(msg)

    @exceptions_to_xmpp_errors
    async def on_legacy_message(self, msg: Message) -> None:
        """
        Meant to be called from :class:`BaseGateway` only.

        :param msg:
        :return:
        """
        # we MUST not use `if m["replace"]["id"]` because it adds the tag if not
        # present. this is a problem for MUC echoed messages
        if msg.get_plugin("replace", check=True) is not None:
            # ignore last message correction (handled by a specific method)
            return
        if msg.get_plugin("apply_to", check=True) is not None:
            # ignore message retraction (handled by a specific method)
            return
        if msg.get_plugin("reactions", check=True) is not None:
            # ignore message reaction fallback.
            # the reaction itself is handled by self.react_from_msg().
            return
        if msg.get_plugin("retract", check=True) is not None:
            # ignore message retraction fallback.
            # the retraction itself is handled by self.on_retract
            return
        if msg.xml.find(".//{*}encrypted") is not None:
            raise XMPPError(
                "bad-request", "You cannot send encrypted messages through this gateway"
            )
        cid = None
        if msg.get_plugin("html", check=True) is not None:
            body = ElementTree.fromstring("<body>" + msg["html"].get_body() + "</body>")
            p = body.findall("p")
            if p is not None and len(p) == 1:
                if p[0].text is None or not p[0].text.strip():
                    images = p[0].findall("img")
                    if len(images) == 1:
                        # no text, single img ⇒ this is a sticker
                        # other cases should be interpreted as "custom emojis" in text
                        src = images[0].get("src")
                        if src is not None and src.startswith("cid:"):
                            cid = src.removeprefix("cid:")

        session, recipient, thread = await self._get_session_recipient_thread(msg)

        if msg.get_plugin("oob", check=True) is not None:
            url = msg["oob"]["url"]
        elif (
            "reference" in msg
            and "sims" in msg["reference"]
            and "sources" in msg["reference"]["sims"]
        ):
            for source in msg["reference"]["sims"]["sources"]["substanzas"]:
                if source["uri"].startswith("http"):
                    url = source["uri"]
                    break
            else:
                url = None
        else:
            url = None

        if msg.get_plugin("reply", check=True):
            text, reply_to_msg_id, reply_to, reply_fallback = await self.__get_reply(
                msg, session, recipient
            )
        else:
            text = msg["body"]
            reply_to_msg_id = None
            reply_to = None
            reply_fallback = None

        if msg.get_plugin("link_previews", check=True):
            link_previews = [
                dict_to_named_tuple(p, LinkPreview) for p in msg["link_previews"]
            ]
        else:
            link_previews = []

        if url:
            legacy_msg_id = await self.__send_url(
                url,
                session,
                recipient,
                reply_to_msg_id=reply_to_msg_id,
                reply_to_fallback_text=reply_fallback,
                reply_to=reply_to,
                thread=thread,
            )
        elif cid:
            legacy_msg_id = await self.__send_bob(
                msg.get_from(),
                cid,
                session,
                recipient,
                reply_to_msg_id=reply_to_msg_id,
                reply_to_fallback_text=reply_fallback,
                reply_to=reply_to,
                thread=thread,
            )
        elif text:
            if isinstance(recipient, LegacyMUC):
                mentions = {"mentions": await recipient.parse_mentions(text)}
            else:
                mentions = {}
            legacy_msg_id = await session.on_text(
                recipient,
                text,
                reply_to_msg_id=reply_to_msg_id,
                reply_to_fallback_text=reply_fallback,
                reply_to=reply_to,
                thread=thread,
                link_previews=link_previews,
                **mentions,
            )
        else:
            log.debug("Ignoring %s", msg.get_id())
            return

        if isinstance(recipient, LegacyMUC):
            stanza_id = await recipient.echo(msg, legacy_msg_id)
        else:
            stanza_id = None
            self.__ack(msg)

        if legacy_msg_id is None:
            return

        with self.xmpp.store.session() as orm:
            if recipient.is_group:
                self.xmpp.store.id_map.set_origin(
                    orm, recipient.stored.id, str(legacy_msg_id), msg.get_id()
                )
                assert stanza_id is not None
                self.xmpp.store.id_map.set_msg(
                    orm,
                    recipient.stored.id,
                    str(legacy_msg_id),
                    [stanza_id],
                    True,
                )
            else:
                self.xmpp.store.id_map.set_msg(
                    orm,
                    recipient.stored.id,
                    str(legacy_msg_id),
                    [msg.get_id()],
                    False,
                )
            if session.MESSAGE_IDS_ARE_THREAD_IDS and (t := msg["thread"]):
                self.xmpp.store.id_map.set_thread(
                    orm, recipient.stored.id, t, str(legacy_msg_id), recipient.is_group
                )
            orm.commit()

    @exceptions_to_xmpp_errors
    async def on_message_correction(self, msg: Message) -> None:
        if msg.get_plugin("retract", check=True) is not None:
            # ignore message retraction fallback (fallback=last msg correction)
            return
        session, recipient, thread = await self._get_session_recipient_thread(msg)
        legacy_id = self._xmpp_msg_id_to_legacy(
            session, msg["replace"]["id"], recipient, True
        )

        if isinstance(recipient, LegacyMUC):
            mentions = await recipient.parse_mentions(msg["body"])
        else:
            mentions = None

        if previews := msg["link_previews"]:
            link_previews = [dict_to_named_tuple(p, LinkPreview) for p in previews]
        else:
            link_previews = []

        if legacy_id is None:
            log.debug("Did not find legacy ID to correct")
            new_legacy_msg_id = await session.on_text(
                recipient,
                "Correction:" + msg["body"],
                thread=thread,
                mentions=mentions,
                link_previews=link_previews,
            )
        elif not msg["body"].strip() and recipient.RETRACTION:
            await session.on_retract(recipient, legacy_id, thread=thread)
            new_legacy_msg_id = None
        elif recipient.CORRECTION:
            new_legacy_msg_id = await session.on_correct(
                recipient,
                msg["body"],
                legacy_id,
                thread=thread,
                mentions=mentions,
                link_previews=link_previews,
            )
        else:
            session.send_gateway_message(
                "Last message correction is not supported by this legacy service. "
                "Slidge will send your correction as new message."
            )
            if recipient.RETRACTION and legacy_id is not None:
                if legacy_id is not None:
                    session.send_gateway_message(
                        "Slidge will attempt to retract the original message you wanted"
                        " to edit."
                    )
                    await session.on_retract(recipient, legacy_id, thread=thread)

            new_legacy_msg_id = await session.on_text(
                recipient,
                "Correction: " + msg["body"],
                thread=thread,
                mentions=mentions,
                link_previews=link_previews,
            )

        if isinstance(recipient, LegacyMUC):
            await recipient.echo(msg, new_legacy_msg_id)
        else:
            self.__ack(msg)
        if new_legacy_msg_id is None:
            return
        with self.xmpp.store.session() as orm:
            self.xmpp.store.id_map.set_msg(
                orm,
                recipient.stored.id,
                new_legacy_msg_id,
                [msg.get_id()],
                recipient.is_group,
            )
            orm.commit()

    @exceptions_to_xmpp_errors
    async def on_message_retract(self, msg: Message):
        session, recipient, thread = await self._get_session_recipient_thread(msg)
        if not recipient.RETRACTION:
            raise XMPPError(
                "bad-request",
                "This legacy service does not support message retraction.",
            )
        xmpp_id: str = msg["retract"]["id"]
        legacy_id = self._xmpp_msg_id_to_legacy(
            session, xmpp_id, recipient, origin=True
        )
        await session.on_retract(recipient, legacy_id, thread=thread)
        if isinstance(recipient, LegacyMUC):
            await recipient.echo(msg, None)
        self.__ack(msg)

    @exceptions_to_xmpp_errors
    async def on_reactions(self, msg: Message):
        session, recipient, thread = await self._get_session_recipient_thread(msg)
        react_to: str = msg["reactions"]["id"]

        special_msg = session.SPECIAL_MSG_ID_PREFIX and react_to.startswith(
            session.SPECIAL_MSG_ID_PREFIX
        )

        if special_msg:
            legacy_id = react_to
        else:
            legacy_id = self._xmpp_msg_id_to_legacy(session, react_to, recipient)

        if not legacy_id:
            log.debug("Ignored reaction from user")
            raise XMPPError(
                "internal-server-error",
                "Could not convert the XMPP msg ID to a legacy ID",
            )

        emojis = [
            remove_emoji_variation_selector_16(r["value"]) for r in msg["reactions"]
        ]
        error_msg = None
        recipient = recipient

        if not special_msg:
            if recipient.REACTIONS_SINGLE_EMOJI and len(emojis) > 1:
                error_msg = "Maximum 1 emoji/message"

            if not error_msg and (
                subset := await recipient.available_emojis(legacy_id)
            ):
                if not set(emojis).issubset(subset):
                    error_msg = f"You can only react with the following emojis: {''.join(subset)}"

        if error_msg:
            session.send_gateway_message(error_msg)
            if not isinstance(recipient, LegacyMUC):
                # no need to carbon for groups, we just don't echo the stanza
                recipient.react(legacy_id, carbon=True)
            await session.on_react(recipient, legacy_id, [], thread=thread)
            raise XMPPError(
                "policy-violation",
                text=error_msg,
                clear=False,
            )

        await session.on_react(recipient, legacy_id, emojis, thread=thread)
        if isinstance(recipient, LegacyMUC):
            await recipient.echo(msg, None)
        else:
            self.__ack(msg)

        with self.xmpp.store.session() as orm:
            multi = self.xmpp.store.id_map.get_xmpp(
                orm, recipient.stored.id, legacy_id, recipient.is_group
            )
        if not multi:
            return
        multi = [m for m in multi if react_to != m]

        if isinstance(recipient, LegacyMUC):
            for xmpp_id in multi:
                mc = copy(msg)
                mc["reactions"]["id"] = xmpp_id
                await recipient.echo(mc)
        elif isinstance(recipient, LegacyContact):
            for xmpp_id in multi:
                recipient.react(legacy_id, emojis, xmpp_id=xmpp_id, carbon=True)

    def __ack(self, msg: Message) -> None:
        if not self.xmpp.PROPER_RECEIPTS:
            self.xmpp.delivery_receipt.ack(msg)

    async def __get_reply(
        self, msg: Message, session: BaseSession, recipient: Recipient
    ) -> tuple[
        str, str | int | None, LegacyContact | LegacyParticipant | None, str | None
    ]:
        try:
            reply_to_msg_id = self._xmpp_msg_id_to_legacy(
                session, msg["reply"]["id"], recipient
            )
        except XMPPError:
            session.log.debug(
                "Could not determine reply-to legacy msg ID, sending quote instead."
            )
            return redact_url(msg["body"]), None, None, None

        reply_to_jid = JID(msg["reply"]["to"])
        reply_to = None
        if msg["type"] == "chat":
            if reply_to_jid.bare != session.user_jid.bare:
                try:
                    reply_to = await session.contacts.by_jid(reply_to_jid)
                except XMPPError:
                    pass
        elif msg["type"] == "groupchat":
            nick = reply_to_jid.resource
            try:
                muc = await session.bookmarks.by_jid(reply_to_jid)
            except XMPPError:
                pass
            else:
                if nick == muc.user_nick:
                    reply_to = await muc.get_user_participant()
                elif not nick:
                    reply_to = muc.get_system_participant()
                else:
                    reply_to = await muc.get_participant(nick, store=False)

        if msg.get_plugin("fallback", check=True) and (
            isinstance(recipient, LegacyMUC) or recipient.REPLIES
        ):
            text = msg["fallback"].get_stripped_body(self.xmpp["xep_0461"].namespace)
            try:
                reply_fallback = redact_url(msg["reply"].get_fallback_body())
            except AttributeError:
                reply_fallback = None
        else:
            text = msg["body"]
            reply_fallback = None

        return text, reply_to_msg_id, reply_to, reply_fallback

    async def __send_url(
        self, url: str, session: BaseSession, recipient: Recipient, **kwargs
    ) -> int | str | None:
        async with self.xmpp.http.get(url) as response:
            if response.status >= 400:
                session.log.warning(
                    "OOB url cannot be downloaded: %s, sending the URL as text"
                    " instead.",
                    response,
                )
                return await session.on_text(recipient, url, **kwargs)

            return await session.on_file(
                recipient, url, http_response=response, **kwargs
            )

    async def __send_bob(
        self, from_: JID, cid: str, session: BaseSession, recipient: Recipient, **kwargs
    ) -> int | str | None:
        with self.xmpp.store.session() as orm:
            sticker = self.xmpp.store.bob.get_sticker(orm, cid)
        if sticker is None:
            await self.xmpp.plugin["xep_0231"].get_bob(
                from_, cid, ifrom=self.xmpp.boundjid
            )
            with self.xmpp.store.session() as orm:
                sticker = self.xmpp.store.bob.get_sticker(orm, cid)
        assert sticker is not None
        return await session.on_sticker(recipient, sticker, **kwargs)


def redact_url(text: str) -> str:
    needle = config.NO_UPLOAD_URL_PREFIX or config.UPLOAD_URL_PREFIX
    if not needle:
        return text
    return text.replace(needle, "")


log = logging.getLogger(__name__)



"""
This module implements vector animated stickers in the lottie format to webp images.


"""

import asyncio
import logging
import warnings
from pathlib import Path

import aiohttp

try:
    import rlottie_python  # type: ignore
except ImportError:
    rlottie_python = None  # type: ignore

from ..core import config
from .types import LegacyAttachment


async def from_url(
    url: str, sticker_id: str, http: aiohttp.ClientSession
) -> LegacyAttachment:
    """
    Get a webp attachment from a URL.

    :param url: URL where the lottie sticker can be downloaded.
    :param sticker_id: A unique identifier for this sticker.
    :param http: The aiohttp.ClientSession used to download the sticker.

    :return: A `LegacyAttachment` with the sticker in the webp format if
        `config.CONVERT_STICKERS == True`, in the original lottie format
        otherwise.
    """
    if not config.CONVERT_STICKERS:
        return _attachment(sticker_id, url=url)
    lottie_path = sticker_path(sticker_id).with_suffix(".json")
    await _download(lottie_path, url, http)
    webp_path = sticker_path(sticker_id).with_suffix(".webp")
    return await convert(lottie_path, webp_path, sticker_id)


async def from_path(path: Path, sticker_id: str) -> LegacyAttachment:
    """
    Get a webp attachment from a path.

    :param path: path to the lottie sticker file.
    :param sticker_id: A unique identifier for this sticker.

    :return: A `LegacyAttachment` with the sticker in the webp format if
        `config.CONVERT_STICKERS == True`, in the original lottie format
        otherwise.
    """
    if not config.CONVERT_STICKERS:
        return _attachment(sticker_id, path=path)
    out_path = sticker_path(sticker_id).with_suffix(".webp")
    return await convert(path, out_path, sticker_id)


def sticker_path(sticker_id="") -> Path:
    """
    Get the path where a sticker is meant be stored

    :param sticker_id: A unique ID for this sticker
    :return: A `path`. It has no suffix, so you might want to use `Path.with_suffix()`
        before writing into it.
    """
    root = config.HOME_DIR / "lottie"
    root.mkdir(exist_ok=True)
    return root / sticker_id


def _attachment(
    sticker_id: str, path: Path | None = None, url: str | None = None
) -> LegacyAttachment:
    if path is not None and path.suffix == ".webp":
        content_type = "image/webp"
    else:
        content_type = None
    return LegacyAttachment(
        url=url,
        path=path,
        legacy_file_id="lottie-" + sticker_id,
        disposition="inline",
        content_type=content_type,
    )


async def _download(path: Path, url: str, http: aiohttp.ClientSession) -> None:
    async with _sticker_download_lock:
        if path.exists():
            return
        with path.open("wb") as fp:
            try:
                resp = await http.get(url, chunked=True)
                async for chunk in resp.content:
                    fp.write(chunk)
            except ValueError as e:
                if e.args[0] != "Chunk too big":
                    raise
                # not sure why this happens but it does sometimes
                resp = await http.get(url)
                fp.write(await resp.content.read())


async def convert(
    input_path: Path,
    output_path: Path,
    sticker_id: str,
    width: int = 256,
    height: int = 256,
) -> LegacyAttachment:
    async with _sticker_conversion_lock:
        if not output_path.exists():
            if rlottie_python is None:
                warnings.warn(
                    "Cannot convert stickers, rlottie-python is not available."
                )
                return _attachment(sticker_id, path=input_path)

            log.debug("Converting sticker %s to video", output_path.stem)
            if input_path.suffix == ".json":
                animation = rlottie_python.LottieAnimation.from_file(str(input_path))
            else:
                animation = rlottie_python.LottieAnimation.from_tgs(str(input_path))
            animation.save_animation(str(output_path), width=width, height=height)

        return _attachment(sticker_id, path=output_path)


async def _main():
    # small entrypoint to easily test that it works
    import sys

    source, destination = sys.argv[1:]
    config.CONVERT_STICKERS = True
    await convert(Path(source), Path(destination), "")


_sticker_conversion_lock: asyncio.Lock = asyncio.Lock()
_sticker_download_lock: asyncio.Lock = asyncio.Lock()


log = logging.getLogger(__name__)


__all__ = ("from_url",)

if __name__ == "__main__":
    asyncio.run(_main())



import unittest

from slixmpp.test import SlixTest

from slidge.slixfix.link_preview import stanza


class TestLinkPreview(SlixTest):
    def setUp(self):
        stanza.register_plugin()

    def testSetLinkPreview(self):
        msg = self.Message()
        msg["link_preview"]["about"] = "https://the.link.example.com/what-was-linked-to"
        msg["link_preview"]["title"] = "A cool title"
        self.check(
            msg,  # language=xml
            """
            <message xmlns="jabber:client">
              <Description xmlns="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
                           about="https://the.link.example.com/what-was-linked-to">
                <title xmlns="https://ogp.me/ns#">A cool title</title>
              </Description>
            </message>
            """,
            use_values=False,
        )
        assert msg["link_preview"]["description"] is None

    def testGetLinkPreview(self):
        # language=xml
        xml_str = """
            <message to="whoever@example.com">
              <body>I wanted to mention https://the.link.example.com/what-was-linked-to</body>
              <rdf:Description xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
                               xmlns:og="https://ogp.me/ns#"
                               rdf:about="https://the.link.example.com/what-was-linked-to">
                <og:title>Page Title</og:title>
                <og:description>Page Description</og:description>
                <og:url>Canonical URL</og:url>
                <og:image>https://link.to.example.com/image.png</og:image>
                <og:type>website</og:type>
                <og:site_name>Some Website</og:site_name>
              </rdf:Description>
            </message>
            """
        xml = self.parse_xml(xml_str)
        msg = self.Message(xml)
        assert msg["link_preview"]["title"] == "Page Title"
        assert msg["link_preview"]["description"] == "Page Description"
        assert msg["link_preview"]["url"] == "Canonical URL"
        assert msg["link_preview"]["image"] == "https://link.to.example.com/image.png"
        assert msg["link_preview"]["site_name"] == "Some Website"
        assert msg["link_preview"]["type"] == "website"
        assert (
            msg["link_preview"]["about"]
            == "https://the.link.example.com/what-was-linked-to"
        )

    def testGetLinkPreviews(self):
        # language=xml
        xml_str = """
            <message to="whoever@example.com">
              <body>I wanted to mention https://the.link.example.com/what-was-linked-to</body>
              <rdf:Description xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
                               xmlns:og="https://ogp.me/ns#"
                               rdf:about="https://the.link.example.com/what-was-linked-to 1">
                <og:title>Page Title 1</og:title>
                <og:description>Page Description 1</og:description>
                <og:url>Canonical URL 1</og:url>
                <og:image>https://link.to.example.com/image.png 1</og:image>
                <og:type>website 1</og:type>
                <og:site_name>Some Website 1</og:site_name>
              </rdf:Description>
              <rdf:Description xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
                               xmlns:og="https://ogp.me/ns#"
                               rdf:about="https://the.link.example.com/what-was-linked-to 2">
                <og:title>Page Title 2</og:title>
                <og:description>Page Description 2</og:description>
                <og:url>Canonical URL 2</og:url>
                <og:image>https://link.to.example.com/image.png 2</og:image>
                <og:type>website 2</og:type>
                <og:site_name>Some Website 2</og:site_name>
              </rdf:Description>
            </message>
            """
        xml = self.parse_xml(xml_str)
        msg = self.Message(xml)
        previews = msg["link_previews"]
        assert len(previews) == 2
        for i, p in enumerate(previews, start=1):
            assert p["title"] == f"Page Title {i}"
            assert p["description"] == f"Page Description {i}"
            assert p["url"] == f"Canonical URL {i}"
            assert p["image"] == f"https://link.to.example.com/image.png {i}"
            assert p["site_name"] == f"Some Website {i}"
            assert p["type"] == f"website {i}"
            assert p["about"] == f"https://the.link.example.com/what-was-linked-to {i}"


suite = unittest.TestLoader().loadTestsFromTestCase(TestLinkPreview)



"""Add origin group ID table

Revision ID: 259a1b71eadf
Revises: cef02a8b1451
Create Date: 2025-11-26 06:50:04.509013

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "259a1b71eadf"
down_revision: str | None = "cef02a8b1451"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    op.create_table(
        "group_msg_origin",
        sa.Column("foreign_key", sa.Integer(), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("legacy_id", sa.String(), nullable=False),
        sa.Column("xmpp_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["foreign_key"],
            ["room.id"],
            name=op.f("fk_group_msg_origin_foreign_key_room"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_group_msg_origin")),
    )
    with op.batch_alter_table("group_msg_origin", schema=None) as batch_op:
        batch_op.create_index(
            "ix_group_msg_origin_legacy_id", ["legacy_id", "foreign_key"], unique=False
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    with op.batch_alter_table("group_msg_origin", schema=None) as batch_op:
        batch_op.drop_index("ix_group_msg_origin_legacy_id")

    op.drop_table("group_msg_origin")
    # ### end Alembic commands ###



import pytest
from conftest import AvatarFixtureMixin

from slidge import BaseGateway, BaseSession
from slidge.contact import LegacyContact
from slidge.core.session import _sessions
from slidge.util.test import SlidgeTest


class Gateway(BaseGateway):
    COMPONENT_NAME = "A test"


class Session(BaseSession):
    async def login(self):
        return "YUP"


class Contact(LegacyContact):
    async def update_info(self):
        if self.legacy_id == "has-vcard":
            self.set_vcard(full_name="A full name", pronouns="they/fae")

    async def fetch_vcard(self):
        if self.legacy_id == "on-demand":
            self.set_vcard(full_name="Lazy")


@pytest.mark.usefixtures("avatar")
class TestSession(AvatarFixtureMixin, SlidgeTest):
    plugin = globals()
    xmpp: Gateway

    def setUp(self):
        super().setUp()
        self.setup_logged_session()

    def _assert_broadcast_on_demand(self):
        self.send(  # language=XML
            """
            <message type="headline"
                     from="on-demand@aim.shakespeare.lit"
                     to="romeo@montague.lit">
              <event xmlns="http://jabber.org/protocol/pubsub#event">
                <items node="urn:xmpp:vcard4">
                  <item id="current"
                        node="urn:xmpp:vcard4">
                    <vcard xmlns="urn:ietf:params:xml:ns:vcard-4.0">
                      <impp>
                        <uri>xmpp:on-demand@aim.shakespeare.lit</uri>
                      </impp>
                      <fn>
                        <text>Lazy</text>
                      </fn>
                    </vcard>
                  </item>
                </items>
              </event>
            </message>
            """
        )

    def test_vcard_in_update_info(self):
        self.run_coro(self.romeo.contacts.by_legacy_id("has-vcard"))
        self.send(  # language=XML
            """
            <message type="headline"
                     from="has-vcard@aim.shakespeare.lit"
                     to="romeo@montague.lit">
              <event xmlns="http://jabber.org/protocol/pubsub#event">
                <items node="urn:xmpp:vcard4">
                  <item id="current"
                        node="urn:xmpp:vcard4">
                    <vcard xmlns="urn:ietf:params:xml:ns:vcard-4.0">
                      <impp>
                        <uri>xmpp:has-vcard@aim.shakespeare.lit</uri>
                      </impp>
                      <fn>
                        <text>A full name</text>
                      </fn>
                      <pronouns>
                        <text>they/fae</text>
                      </pronouns>
                    </vcard>
                  </item>
                </items>
              </event>
            </message>
            """
        )
        assert self.next_sent() is None

    def test_vcard_outside_update_info(self):
        self.juliet.set_vcard("Another full name")
        self.send(  # language=XML
            """
            <message type="headline"
                     from="juliet@aim.shakespeare.lit"
                     to="romeo@montague.lit">
              <event xmlns="http://jabber.org/protocol/pubsub#event">
                <items node="urn:xmpp:vcard4">
                  <item id="current"
                        node="urn:xmpp:vcard4">
                    <vcard xmlns="urn:ietf:params:xml:ns:vcard-4.0">
                      <impp>
                        <uri>xmpp:juliet@aim.shakespeare.lit</uri>
                      </impp>
                      <fn>
                        <text>Another full name</text>
                      </fn>
                    </vcard>
                  </item>
                </items>
              </event>
            </message>
            """
        )
        assert self.next_sent() is None

    def test_fetch_raw_iq(self):
        self.run_coro(self.romeo.contacts.by_legacy_id("on-demand"))
        assert self.next_sent() is None
        self.recv(  # language=XML
            """
            <iq from='romeo@montague.lit'
                id='fetch'
                to='on-demand@aim.shakespeare.lit'
                type='get'>
              <vcard xmlns='urn:ietf:params:xml:ns:vcard-4.0' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq from="on-demand@aim.shakespeare.lit"
                id="fetch"
                to="romeo@montague.lit"
                type="result">
              <vcard xmlns="urn:ietf:params:xml:ns:vcard-4.0">
                <impp>
                  <uri>xmpp:on-demand@aim.shakespeare.lit</uri>
                </impp>
                <fn>
                  <text>Lazy</text>
                </fn>
              </vcard>
            </iq>
            """,
            use_values=False,
        )
        self._assert_broadcast_on_demand()
        assert self.next_sent() is None

    def test_fetch_raw_iq_empty(self):
        self.recv(  # language=XML
            """
            <iq from='romeo@montague.lit'
                id='fetch'
                to='juliet@aim.shakespeare.lit'
                type='get'>
              <vcard xmlns='urn:ietf:params:xml:ns:vcard-4.0' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq from="juliet@aim.shakespeare.lit"
                id="fetch"
                to="romeo@montague.lit"
                type="result">
              <vcard xmlns="urn:ietf:params:xml:ns:vcard-4.0" />
            </iq>
            """,
            use_values=False,
        )
        assert self.next_sent() is None

    def test_fetch_pubsub(self):
        self.recv(  # language=XML
            """
            <iq from='romeo@montague.lit'
                id='fetch'
                to='on-demand@aim.shakespeare.lit'
                type='get'>
              <pubsub xmlns='http://jabber.org/protocol/pubsub'>
                <items node='urn:xmpp:vcard4' />
              </pubsub>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq from="on-demand@aim.shakespeare.lit"
                id="fetch"
                to="romeo@montague.lit"
                type="result">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <items node="urn:xmpp:vcard4">
                  <item id="current">
                    <vcard xmlns="urn:ietf:params:xml:ns:vcard-4.0">
                      <impp>
                        <uri>xmpp:on-demand@aim.shakespeare.lit</uri>
                      </impp>
                      <fn>
                        <text>Lazy</text>
                      </fn>
                    </vcard>
                  </item>
                </items>
              </pubsub>
            </iq>
            """
        )
        self._assert_broadcast_on_demand()
        assert self.next_sent() is None

    def test_fetch_pubsub_empty(self):
        self.recv(  # language=XML
            """
            <iq from='romeo@montague.lit'
                id='fetch'
                to='juliet@aim.shakespeare.lit'
                type='get'>
              <pubsub xmlns='http://jabber.org/protocol/pubsub'>
                <items node='urn:xmpp:vcard4' />
              </pubsub>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq from="juliet@aim.shakespeare.lit"
                id="fetch"
                to="romeo@montague.lit"
                type="result">
              <error type="cancel">
                <item-not-found xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
              </error>
            </iq>
            """
        )
        assert self.next_sent() is None



from slixmpp import JID, CoroutineCallback, Iq, StanzaPath
from slixmpp.exceptions import XMPPError
from slixmpp.xmlstream import StanzaBase

from ..util import DispatcherMixin, exceptions_to_xmpp_errors


class MucAdminMixin(DispatcherMixin):
    __slots__: list[str] = []

    def __init__(self, xmpp) -> None:
        super().__init__(xmpp)
        self.xmpp.register_handler(
            CoroutineCallback(
                "MUCModerate",
                StanzaPath("iq/moderate"),
                self.on_user_moderation,
            )
        )
        self.xmpp.register_handler(
            CoroutineCallback(
                "MUCSetAffiliation",
                StanzaPath("iq@type=set/mucadmin_query"),
                self.on_user_set_affiliation,
            )
        )
        self.xmpp.register_handler(
            CoroutineCallback(
                "MUCGetAffiliation",
                StanzaPath("iq@type=get/mucadmin_query"),
                self.on_muc_admin_query_get,
            )
        )

    @exceptions_to_xmpp_errors
    async def on_user_moderation(self, iq: StanzaBase) -> None:
        assert isinstance(iq, Iq)
        muc = await self.get_muc_from_stanza(iq)

        moderate = iq["moderate"]
        xmpp_id = iq["moderate"]["id"]
        if not xmpp_id:
            raise XMPPError("bad-request", "Missing moderated message ID")

        if not moderate["retract"]:
            raise XMPPError(
                "feature-not-implemented",
                "Slidge only implements moderation/retraction",
            )

        legacy_id = self._xmpp_msg_id_to_legacy(muc.session, xmpp_id, muc)
        await muc.session.on_moderate(muc, legacy_id, moderate["reason"] or None)
        iq.reply(clear=True).send()

    @exceptions_to_xmpp_errors
    async def on_user_set_affiliation(self, iq: StanzaBase) -> None:
        assert isinstance(iq, Iq)
        muc = await self.get_muc_from_stanza(iq)

        item = iq["mucadmin_query"]["item"]
        if item["jid"]:
            contact = await muc.session.contacts.by_jid(JID(item["jid"]))
        else:
            part = await muc.get_participant(
                item["nick"], fill_first=True, create=False
            )
            if part is None:
                raise XMPPError("item-not-found")
            assert part.contact is not None
            contact = part.contact

        if item["affiliation"]:
            await muc.on_set_affiliation(
                contact,
                item["affiliation"],
                item["reason"] or None,
                item["nick"] or None,
            )
        elif item["role"] == "none":
            await muc.on_kick(contact, item["reason"] or None)

        iq.reply(clear=True).send()

    @exceptions_to_xmpp_errors
    async def on_muc_admin_query_get(self, iq: StanzaBase) -> None:
        assert isinstance(iq, Iq)
        affiliation = iq["mucadmin_query"]["item"]["affiliation"]

        if not affiliation:
            raise XMPPError("bad-request")

        session = await self._get_session(iq, 1, logged=True)
        muc = await session.bookmarks.by_jid(iq.get_to())

        reply = iq.reply()
        reply.enable("mucadmin_query")
        async for participant in muc.get_participants(affiliation):
            reply["mucadmin_query"].append(participant.mucadmin_item())
        reply.send()



import logging

from slixmpp import JID, Presence
from slixmpp.exceptions import XMPPError

from ...contact.roster import ContactIsUser
from ...util.types import AnyBaseSession
from ...util.util import merge_resources
from ..session import BaseSession
from .util import DispatcherMixin, exceptions_to_xmpp_errors


class _IsDirectedAtComponent(Exception):
    def __init__(self, session: BaseSession) -> None:
        self.session = session


class PresenceHandlerMixin(DispatcherMixin):
    __slots__: list[str] = []

    def __init__(self, xmpp) -> None:
        super().__init__(xmpp)

        xmpp.add_event_handler("presence_subscribe", self._handle_subscribe)
        xmpp.add_event_handler("presence_subscribed", self._handle_subscribed)
        xmpp.add_event_handler("presence_unsubscribe", self._handle_unsubscribe)
        xmpp.add_event_handler("presence_unsubscribed", self._handle_unsubscribed)
        xmpp.add_event_handler("presence_probe", self._handle_probe)
        xmpp.add_event_handler("presence", self.on_presence)

    async def __get_contact(self, pres: Presence):
        sess = await self._get_session(pres)
        pto = pres.get_to()
        if pto == self.xmpp.boundjid.bare:
            raise _IsDirectedAtComponent(sess)
        await sess.contacts.ready
        return await sess.contacts.by_jid(pto)

    @exceptions_to_xmpp_errors
    async def _handle_subscribe(self, pres: Presence) -> None:
        try:
            contact = await self.__get_contact(pres)
        except _IsDirectedAtComponent:
            pres.reply().send()
            return

        if contact.is_friend:
            pres.reply().send()
        else:
            await contact.on_friend_request(pres["status"])

    @exceptions_to_xmpp_errors
    async def _handle_unsubscribe(self, pres: Presence) -> None:
        pres.reply().send()

        try:
            contact = await self.__get_contact(pres)
        except _IsDirectedAtComponent as e:
            e.session.send_gateway_message("Bye bye!")
            await e.session.kill_by_jid(e.session.user_jid)
            return

        contact.is_friend = False
        await contact.on_friend_delete(pres["status"])

    @exceptions_to_xmpp_errors
    async def _handle_subscribed(self, pres: Presence) -> None:
        try:
            contact = await self.__get_contact(pres)
        except _IsDirectedAtComponent:
            return

        await contact.on_friend_accept()
        contact.send_last_presence(force=True)

    @exceptions_to_xmpp_errors
    async def _handle_unsubscribed(self, pres: Presence) -> None:
        try:
            contact = await self.__get_contact(pres)
        except _IsDirectedAtComponent:
            return

        if contact.is_friend:
            contact.is_friend = False
            await contact.on_friend_delete(pres["status"])

    @exceptions_to_xmpp_errors
    async def _handle_probe(self, pres: Presence) -> None:
        try:
            contact = await self.__get_contact(pres)
        except _IsDirectedAtComponent:
            session = await self._get_session(pres)
            session.send_cached_presence(pres.get_from())
            return
        if contact.is_friend:
            contact.send_last_presence(force=True)
        else:
            reply = pres.reply()
            reply["type"] = "unsubscribed"
            reply.send()

    @exceptions_to_xmpp_errors
    async def on_presence(self, p: Presence) -> None:
        if p.get_plugin("muc_join", check=True):
            # handled in on_groupchat_join
            # without this early return, since we switch from and to in this
            # presence stanza, on_groupchat_join ends up trying to instantiate
            # a MUC with the user's JID, which in turn leads to slidge sending
            # a (error) presence from=the user's JID, which terminates the
            # XML stream.
            return

        session = await self._get_session(p)

        pto = p.get_to()
        if pto == self.xmpp.boundjid.bare:
            await self._on_presence_to_component(session, p)
            return

        if p.get_type() == "available":
            try:
                contact = await session.contacts.by_jid(pto)
            except XMPPError:
                contact = None
            except ContactIsUser:
                raise XMPPError(
                    "bad-request", "Actions with yourself are not supported."
                )
            if contact is not None:
                await self.xmpp.pubsub.on_presence_available(p, contact)
                return

        muc = session.bookmarks.by_jid_only_if_exists(JID(pto.bare))

        if muc is not None and p.get_type() == "unavailable":
            return muc.on_presence_unavailable(p)

        if muc is None or p.get_from().resource not in muc.get_user_resources():
            return

        if pto.resource == muc.user_nick:
            # Ignore presence stanzas with the valid nick.
            # even if joined to the group, we might receive those from clients,
            # when setting a status message, or going away, etc.
            return

        # We can't use XMPPError here because XMPPError does not have a way to
        # add the <x xmlns="http://jabber.org/protocol/muc" /> element

        error_stanza = p.error()
        error_stanza.set_to(p.get_from())
        error_stanza.set_from(pto)
        error_stanza.enable("muc_join")  # <x xmlns="http://jabber.org/protocol/muc" />
        error_stanza.enable("error")
        error_stanza["error"]["type"] = "cancel"
        error_stanza["error"]["by"] = muc.jid
        error_stanza["error"]["condition"] = "not-acceptable"
        error_stanza["error"]["text"] = (
            "Slidge does not let you change your nickname in groups."
        )
        error_stanza.send()

    async def _on_presence_to_component(
        self, session: AnyBaseSession, p: Presence
    ) -> None:
        session.log.debug("Received a presence from %s", p.get_from())
        if (ptype := p.get_type()) not in _USEFUL_PRESENCES:
            return
        if not session.user.preferences.get("sync_presence", False):
            session.log.debug("User does not want to sync their presence")
            return
        # NB: get_type() returns either a proper presence type or
        #     a presence show if available. Weird, weird, weird slix.
        resources = self.xmpp.roster[self.xmpp.boundjid.bare][p.get_from()].resources
        try:
            await session.on_presence(
                p.get_from().resource,
                ptype,  # type: ignore
                p["status"],
                resources,
                merge_resources(resources),
            )
        except NotImplementedError:
            pass
        if p.get_type() == "available":
            await self.xmpp.pubsub.on_presence_available(p, None)
            for contact in session.contacts:
                await self.xmpp.pubsub.on_presence_available(p, contact)


_USEFUL_PRESENCES = {"available", "unavailable", "away", "chat", "dnd", "xa"}

log = logging.getLogger(__name__)



import logging
from collections.abc import Iterable
from datetime import datetime

from slixmpp import Message

from ...util.archive_msg import HistoryMessage
from ...util.types import (
    LegacyMessageType,
    LegacyThreadType,
    LinkPreview,
    MessageReference,
    ProcessingHint,
)
from ...util.util import add_quote_prefix
from .message_maker import MessageMaker


class TextMessageMixin(MessageMaker):
    def __default_hints(self, hints: Iterable[ProcessingHint] | None = None):
        if hints is not None:
            return hints
        elif self.mtype == "chat":
            return {"markable", "store"}
        elif self.mtype == "groupchat":
            return {"markable"}

    def _replace_id(self, legacy_msg_id: LegacyMessageType) -> str:
        if self.mtype == "groupchat":
            with self.xmpp.store.session() as orm:
                ids = self.xmpp.store.id_map.get_origin(
                    orm, self._recipient_pk(), str(legacy_msg_id)
                )
                if ids:
                    if len(ids) > 1:
                        log.warning(
                            "More than 1 origin msg ID for '%s': '%s'",
                            legacy_msg_id,
                            ids,
                        )
                    return ids[0]
                return self.session.legacy_to_xmpp_msg_id(legacy_msg_id)
        else:
            return self._legacy_to_xmpp(legacy_msg_id)[0]

    def send_text(
        self,
        body: str,
        legacy_msg_id: LegacyMessageType | None = None,
        *,
        when: datetime | None = None,
        reply_to: MessageReference | None = None,
        thread: LegacyThreadType | None = None,
        hints: Iterable[ProcessingHint] | None = None,
        carbon: bool = False,
        archive_only: bool = False,
        correction: bool = False,
        correction_event_id: LegacyMessageType | None = None,
        link_previews: list[LinkPreview] | None = None,
        **send_kwargs,
    ):
        """
        Send a text message from this :term:`XMPP Entity`.

        :param body: Content of the message
        :param legacy_msg_id: If you want to be able to transport read markers from the gateway
            user to the legacy network, specify this
        :param when: when the message was sent, for a "delay" tag (:xep:`0203`)
        :param reply_to: Quote another message (:xep:`0461`)
        :param hints:
        :param thread:
        :param carbon: (only used if called on a :class:`LegacyContact`)
            Set this to ``True`` if this is actually a message sent **to** the
            :class:`LegacyContact` by the :term:`User`.
            Use this to synchronize outgoing history for legacy official apps.
        :param correction: whether this message is a correction or not
        :param correction_event_id: in the case where an ID is associated with the legacy
            'correction event', specify it here to use it on the XMPP side. If not specified,
            a random ID will be used.
        :param link_previews: A little of sender (or server, or gateway)-generated
            previews of URLs linked in the body.
        :param archive_only: (only in groups) Do not send this message to user,
            but store it in the archive. Meant to be used during ``MUC.backfill()``
        """
        if carbon and not self.is_participant:
            with self.xmpp.store.session() as orm:
                if not correction and self.xmpp.store.id_map.was_sent_by_user(
                    orm, self._recipient_pk(), str(legacy_msg_id), False
                ):
                    log.warning(
                        "Carbon message for a message an XMPP has sent? This is a bug! %s",
                        legacy_msg_id,
                    )
                    return
                self.xmpp.store.id_map.set_msg(
                    orm,
                    self._recipient_pk(),
                    str(legacy_msg_id),
                    [self.session.legacy_to_xmpp_msg_id(legacy_msg_id)],
                    False,
                )
                orm.commit()
        hints = self.__default_hints(hints)
        msg = self._make_message(
            mbody=body,
            legacy_msg_id=correction_event_id if correction else legacy_msg_id,
            when=when,
            reply_to=reply_to,
            hints=hints or (),
            carbon=carbon,
            thread=thread,
            link_previews=link_previews,
        )
        if correction:
            msg["replace"]["id"] = self._replace_id(legacy_msg_id)
        return self._send(
            msg,
            archive_only=archive_only,
            carbon=carbon,
            legacy_msg_id=legacy_msg_id,
            **send_kwargs,
        )

    def correct(
        self,
        legacy_msg_id: LegacyMessageType,
        new_text: str,
        *,
        when: datetime | None = None,
        reply_to: MessageReference | None = None,
        thread: LegacyThreadType | None = None,
        hints: Iterable[ProcessingHint] | None = None,
        carbon: bool = False,
        archive_only: bool = False,
        correction_event_id: LegacyMessageType | None = None,
        link_previews: list[LinkPreview] | None = None,
        **send_kwargs,
    ) -> None:
        """
        Modify a message that was previously sent by this :term:`XMPP Entity`.

        Uses last message correction (:xep:`0308`)

        :param new_text: New content of the message
        :param legacy_msg_id: The legacy message ID of the message to correct
        :param when: when the message was sent, for a "delay" tag (:xep:`0203`)
        :param reply_to: Quote another message (:xep:`0461`)
        :param hints:
        :param thread:
        :param carbon: (only in 1:1) Reflect a message sent to this ``Contact`` by the user.
            Use this to synchronize outgoing history for legacy official apps.
        :param archive_only: (only in groups) Do not send this message to user,
            but store it in the archive. Meant to be used during ``MUC.backfill()``
        :param correction_event_id: in the case where an ID is associated with the legacy
            'correction event', specify it here to use it on the XMPP side. If not specified,
            a random ID will be used.
        :param link_previews: A little of sender (or server, or gateway)-generated
            previews of URLs linked in the body.
        """
        self.send_text(
            new_text,
            legacy_msg_id,
            when=when,
            reply_to=reply_to,
            hints=hints,
            carbon=carbon,
            thread=thread,
            correction=True,
            archive_only=archive_only,
            correction_event_id=correction_event_id,
            link_previews=link_previews,
            **send_kwargs,
        )

    def react(
        self,
        legacy_msg_id: LegacyMessageType,
        emojis: Iterable[str] = (),
        thread: LegacyThreadType | None = None,
        **kwargs,
    ) -> None:
        """
        Send a reaction (:xep:`0444`) from this :term:`XMPP Entity`.

        :param legacy_msg_id: The message which the reaction refers to.
        :param emojis: An iterable of emojis used as reactions
        :param thread:
        """
        xmpp_id = kwargs.pop("xmpp_id", None)
        if xmpp_id:
            xmpp_ids = [xmpp_id]
        else:
            xmpp_ids = self._legacy_to_xmpp(legacy_msg_id)
        for xmpp_id in xmpp_ids:
            msg = self._make_message(
                hints={"store"}, carbon=bool(kwargs.get("carbon")), thread=thread
            )
            self.xmpp["xep_0444"].set_reactions(
                msg, to_id=xmpp_id, reactions=set(emojis)
            )
            self.__add_reaction_fallback(msg, legacy_msg_id, emojis)
            self._send(msg, **kwargs)

    def __add_reaction_fallback(
        self,
        msg: Message,
        legacy_msg_id: LegacyMessageType,
        emojis: Iterable[str] = (),
    ) -> None:
        if not self.session.user.preferences.get("reaction_fallback", False):
            return
        msg["fallback"]["for"] = self.xmpp.plugin["xep_0444"].namespace
        msg["fallback"].enable("body")
        msg["body"] = " ".join(emojis)
        if not self.is_participant:
            return
        with self.xmpp.store.session() as orm:
            archived = self.xmpp.store.mam.get_by_legacy_id(
                orm, self.muc.stored.id, str(legacy_msg_id)
            )
            if archived is None:
                return
            history_msg = HistoryMessage(archived.stanza)
            msg["body"] = (
                add_quote_prefix(history_msg.stanza["body"]) + "\n" + msg["body"]
            )

    def retract(
        self,
        legacy_msg_id: LegacyMessageType,
        thread: LegacyThreadType | None = None,
        **kwargs,
    ) -> None:
        """
        Send a message retraction (:XEP:`0424`) from this :term:`XMPP Entity`.

        :param legacy_msg_id: Legacy ID of the message to delete
        :param thread:
        """
        xmpp_ids = self._legacy_to_xmpp(legacy_msg_id)
        replace_id = self._replace_id(legacy_msg_id)
        if replace_id not in xmpp_ids:
            xmpp_ids.append(replace_id)
        for xmpp_id in xmpp_ids:
            msg = self._make_message(
                state=None,
                hints={"store"},
                mbody=f"/me retracted the message {legacy_msg_id}",
                carbon=bool(kwargs.get("carbon")),
                thread=thread,
            )
            msg.enable("fallback")
            # namespace version mismatch between slidge and slixmpp, update me later
            msg["fallback"]["for"] = self.xmpp["xep_0424"].namespace[:-1] + "1"
            msg["retract"]["id"] = msg["replace"]["id"] = xmpp_id
            self._send(msg, **kwargs)


log = logging.getLogger(__name__)



import logging
from typing import TYPE_CHECKING, Any

import sqlalchemy as sa
from slixmpp.exceptions import XMPPError
from slixmpp.plugins.xep_0030.stanza.items import DiscoItems
from slixmpp.types import OptJid

from ...db.models import Room
from .util import DispatcherMixin

if TYPE_CHECKING:
    from slidge.core.gateway import BaseGateway


class DiscoMixin(DispatcherMixin):
    __slots__: list[str] = []

    def __init__(self, xmpp: "BaseGateway") -> None:
        super().__init__(xmpp)

        xmpp.plugin["xep_0030"].set_node_handler(
            "get_info",
            jid=None,
            node=None,
            handler=self.get_info,
        )

        xmpp.plugin["xep_0030"].set_node_handler(
            "get_items",
            jid=None,
            node=None,
            handler=self.get_items,
        )

    async def get_info(self, jid: OptJid, node: str | None, ifrom: OptJid, data: Any):
        if ifrom == self.xmpp.boundjid.bare or jid in (self.xmpp.boundjid.bare, None):
            return self.xmpp.plugin["xep_0030"].static.get_info(jid, node, ifrom, data)

        if ifrom is None:
            raise XMPPError("subscription-required")

        assert jid is not None
        session = await self._get_session_from_jid(jid=ifrom)

        log.debug("Looking for entity: %s", jid)

        entity = await session.get_contact_or_group_or_participant(jid)

        if entity is None:
            raise XMPPError("item-not-found")

        return await entity.get_disco_info(jid, node)

    async def get_items(self, jid: OptJid, node: str | None, ifrom: OptJid, data: Any):
        if ifrom is None:
            raise XMPPError("bad-request")

        if jid != self.xmpp.boundjid.bare:
            return DiscoItems()

        assert ifrom is not None
        session = await self._get_session_from_jid(ifrom)

        d = DiscoItems()
        with self.xmpp.store.session() as orm:
            for room in orm.execute(
                sa.select(Room)
                .options(sa.orm.load_only(Room.jid, Room.name))
                .filter_by(user=session.user)
                .order_by(Room.name)
            ).scalars():
                d.add_item(room.jid, name=room.name)

        return d


log = logging.getLogger(__name__)



import logging
from collections.abc import Iterator
from typing import TYPE_CHECKING, Generic, Literal, overload

from slixmpp import JID
from slixmpp.exceptions import XMPPError

from ..db.models import Room
from ..util import SubclassableOnce
from ..util.jid_escaping import ESCAPE_TABLE, unescape_node
from ..util.lock import NamedLockMixin
from ..util.types import LegacyGroupIdType, LegacyMUCType
from .room import LegacyMUC

if TYPE_CHECKING:
    from slidge.core.session import BaseSession


class LegacyBookmarks(
    Generic[LegacyGroupIdType, LegacyMUCType],
    NamedLockMixin,
    SubclassableOnce,
):
    """
    This is instantiated once per :class:`~slidge.BaseSession`
    """

    _muc_cls: type[LegacyMUCType]

    def __init__(self, session: "BaseSession") -> None:
        self.session = session
        self.xmpp = session.xmpp
        self.user_jid = session.user_jid

        self._user_nick: str = self.session.user_jid.node

        super().__init__()
        self.log = logging.getLogger(f"{self.user_jid.bare}:bookmarks")
        self.ready = self.session.xmpp.loop.create_future()
        if not self.xmpp.GROUPS:
            self.ready.set_result(True)

    @property
    def user_nick(self):
        return self._user_nick

    @user_nick.setter
    def user_nick(self, nick: str) -> None:
        self._user_nick = nick

    def from_store(self, stored: Room) -> LegacyMUCType:
        return self._muc_cls(self.session, stored)

    def __iter__(self) -> Iterator[LegacyMUC]:
        with self.xmpp.store.session() as orm:
            for stored in orm.query(Room).filter_by(user=self.session.user).all():
                if stored.updated:
                    yield self.from_store(stored)

    def __repr__(self) -> str:
        return f"<Bookmarks of {self.user_jid}>"

    async def legacy_id_to_jid_local_part(self, legacy_id: LegacyGroupIdType) -> str:
        return await self.legacy_id_to_jid_username(legacy_id)

    async def jid_local_part_to_legacy_id(self, local_part: str) -> LegacyGroupIdType:
        return await self.jid_username_to_legacy_id(local_part)

    async def legacy_id_to_jid_username(self, legacy_id: LegacyGroupIdType) -> str:
        """
        The default implementation calls ``str()`` on the legacy_id and
        escape characters according to :xep:`0106`.

        You can override this class and implement a more subtle logic to raise
        an :class:`~slixmpp.exceptions.XMPPError` early

        :param legacy_id:
        :return:
        """
        return str(legacy_id).translate(ESCAPE_TABLE)

    async def jid_username_to_legacy_id(self, username: str):
        """

        :param username:
        :return:
        """
        return unescape_node(username)

    async def by_jid(self, jid: JID) -> LegacyMUCType:
        if jid.resource:
            jid = JID(jid.bare)
        async with self.lock(("bare", jid.bare)):
            legacy_id = await self.jid_local_part_to_legacy_id(jid.node)
            if self.get_lock(("legacy_id", legacy_id)):
                self.session.log.debug("Already updating %s via by_legacy_id()", jid)
                return await self.by_legacy_id(legacy_id)

            with self.session.xmpp.store.session() as orm:
                stored = (
                    orm.query(Room)
                    .filter_by(user_account_id=self.session.user_pk, jid=jid)
                    .one_or_none()
                )
            if stored is None:
                stored = Room(
                    user_account_id=self.session.user_pk,
                    jid=jid,
                    legacy_id=legacy_id,
                )
            return await self.__update_if_needed(stored)

    def by_jid_only_if_exists(self, jid: JID) -> LegacyMUCType | None:
        with self.xmpp.store.session() as orm:
            stored = (
                orm.query(Room).filter_by(user=self.session.user, jid=jid).one_or_none()
            )
            if stored is not None and stored.updated:
                return self.from_store(stored)
        return None

    @overload
    async def by_legacy_id(self, legacy_id: LegacyGroupIdType) -> "LegacyMUCType": ...

    @overload
    async def by_legacy_id(
        self, legacy_id: LegacyGroupIdType, create: Literal[False]
    ) -> "LegacyMUCType | None": ...

    @overload
    async def by_legacy_id(
        self, legacy_id: LegacyGroupIdType, create: Literal[True]
    ) -> "LegacyMUCType": ...

    async def by_legacy_id(
        self, legacy_id: LegacyGroupIdType, create: bool = False
    ) -> LegacyMUCType | None:
        async with self.lock(("legacy_id", legacy_id)):
            local = await self.legacy_id_to_jid_local_part(legacy_id)
            jid = JID(f"{local}@{self.xmpp.boundjid}")
            if self.get_lock(("bare", jid.bare)):
                self.session.log.debug("Already updating %s via by_jid()", jid)
                if create:
                    return await self.by_jid(jid)
                else:
                    return self.by_jid_only_if_exists(jid)

            with self.xmpp.store.session() as orm:
                stored = (
                    orm.query(Room)
                    .filter_by(
                        user_account_id=self.session.user_pk,
                        legacy_id=str(legacy_id),
                    )
                    .one_or_none()
                )
            if stored is None:
                stored = Room(
                    user_account_id=self.session.user_pk,
                    jid=jid,
                    legacy_id=str(legacy_id),
                )
            return await self.__update_if_needed(stored)

    async def __update_if_needed(self, stored: Room) -> LegacyMUCType:
        muc = self.from_store(stored)
        if muc.stored.updated:
            return muc

        with muc.updating_info(merge=False or muc._ALL_INFO_FILLED_ON_STARTUP):
            try:
                await muc.update_info()
            except NotImplementedError:
                pass
            except XMPPError:
                raise
            except Exception as e:
                raise XMPPError("internal-server-error", str(e))

        return muc

    async def fill(self):
        """
        Establish a user's known groups.

        This has to be overridden in plugins with group support and at the
        minimum, this should ``await self.by_legacy_id(group_id)`` for all
        the groups a user is part of.

        Slidge internals will call this on successful :meth:`BaseSession.login`

        """
        if self.xmpp.GROUPS:
            raise NotImplementedError(
                "The plugin advertised support for groups but"
                " LegacyBookmarks.fill() was not overridden."
            )

    async def remove(
        self,
        muc: LegacyMUC,
        reason: str = "You left this group from the official client.",
        kick: bool = True,
    ) -> None:
        """
        Delete everything about a specific group.

        This should be called when the user leaves the group from the official
        app.

        :param muc: The MUC to remove.
        :param reason: Optionally, a reason why this group was removed.
        :param kick: Whether the user should be kicked from this group. Set this
            to False in case you do this somewhere else in your code, eg, on
            receiving the confirmation that the group was deleted.
        """
        if kick:
            user_participant = await muc.get_user_participant()
            user_participant.kick(reason)
        with self.xmpp.store.session() as orm:
            orm.add(muc.stored)
            orm.refresh(muc.stored)
            orm.delete(muc.stored)
            orm.commit()



import asyncio
import functools
import inspect
import logging
from collections.abc import Callable
from functools import partial
from typing import TYPE_CHECKING, Any, Optional

from slixmpp import JID, Iq
from slixmpp.exceptions import XMPPError
from slixmpp.plugins.xep_0004 import Form as SlixForm  # type: ignore[attr-defined]
from slixmpp.plugins.xep_0030.stanza.items import DiscoItems

from ..core import config
from ..util.util import strip_leading_emoji
from . import Command, CommandResponseType, Confirmation, Form, TableResult
from .base import FormField
from .categories import CommandCategory

if TYPE_CHECKING:
    from ..core.gateway import BaseGateway
    from ..core.session import BaseSession


AdhocSessionType = dict[str, Any]


class AdhocProvider:
    """
    A slixmpp-like plugin to handle adhoc commands, with less boilerplate and
    untyped dict values than slixmpp.
    """

    FORM_TIMEOUT = 120  # seconds

    def __init__(self, xmpp: "BaseGateway") -> None:
        self.xmpp = xmpp
        self._commands = dict[str, Command]()
        self._categories = dict[str, list[Command]]()
        xmpp.plugin["xep_0030"].set_node_handler(
            "get_items",
            jid=xmpp.boundjid,
            node=self.xmpp.plugin["xep_0050"].stanza.Command.namespace,
            handler=self.get_items,
        )
        self.__timeouts: dict[str, asyncio.TimerHandle] = {}

    async def __wrap_initial_handler(
        self, command: Command, iq: Iq, adhoc_session: AdhocSessionType
    ) -> AdhocSessionType:
        ifrom = iq.get_from()
        session = command.raise_if_not_authorized(ifrom)
        result = await self.__wrap_handler(command.run, session, ifrom)
        return await self.__handle_result(session, result, adhoc_session)

    async def __handle_category_list(
        self, category: CommandCategory, iq: Iq, adhoc_session: AdhocSessionType
    ) -> AdhocSessionType:
        try:
            session = self.xmpp.get_session_from_stanza(iq)
        except XMPPError:
            session = None
        commands: dict[str, Command] = {}
        for command in self._categories[category.node]:
            try:
                command.raise_if_not_authorized(iq.get_from())
            except XMPPError:
                continue
            commands[command.NODE] = command
        if len(commands) == 0:
            raise XMPPError(
                "not-authorized", "There is no command you can run in this category"
            )
        return await self.__handle_result(
            session,
            Form(
                category.name,
                "",
                [
                    FormField(
                        var="command",
                        label="Command",
                        type="list-single",
                        options=[
                            {
                                "label": strip_leading_emoji_if_needed(command.NAME),
                                "value": command.NODE,
                            }
                            for command in commands.values()
                        ],
                    )
                ],
                partial(self.__handle_category_choice, commands),
            ),
            adhoc_session,
        )

    async def __handle_category_choice(
        self,
        commands: dict[str, Command],
        form_values: dict[str, str],
        session: "BaseSession[Any, Any]",
        jid: JID,
    ):
        command = commands[form_values["command"]]
        result = await self.__wrap_handler(command.run, session, jid)
        return result

    async def __handle_result(
        self,
        session: Optional["BaseSession[Any, Any]"],
        result: CommandResponseType,
        adhoc_session: AdhocSessionType,
    ) -> AdhocSessionType:
        if isinstance(result, str) or result is None:
            adhoc_session["has_next"] = False
            adhoc_session["next"] = None
            adhoc_session["payload"] = None
            adhoc_session["notes"] = [("info", result or "Success!")]
            return adhoc_session

        if isinstance(result, Form):
            adhoc_session["next"] = partial(self.__wrap_form_handler, session, result)
            adhoc_session["has_next"] = True
            adhoc_session["payload"] = result.get_xml()
            if result.timeout_handler is not None:
                self.__timeouts[adhoc_session["id"]] = self.xmpp.loop.call_later(
                    self.FORM_TIMEOUT,
                    partial(
                        self.__wrap_timeout, result.timeout_handler, adhoc_session["id"]
                    ),
                )
            return adhoc_session

        if isinstance(result, Confirmation):
            adhoc_session["next"] = partial(self.__wrap_confirmation, session, result)
            adhoc_session["has_next"] = True
            adhoc_session["payload"] = result.get_form()
            adhoc_session["next"] = partial(self.__wrap_confirmation, session, result)
            return adhoc_session

        if isinstance(result, TableResult):
            adhoc_session["next"] = None
            adhoc_session["has_next"] = False
            adhoc_session["payload"] = result.get_xml()
            return adhoc_session

        raise XMPPError("internal-server-error", text="OOPS!")

    def __wrap_timeout(self, handler: Callable[[], None], session_id: str) -> None:
        try:
            del self.xmpp.plugin["xep_0050"].sessions[session_id]
        except KeyError:
            log.error("Timeout but session could not be found: %s", session_id)
            handler()

    @staticmethod
    async def __wrap_handler(f: Callable | functools.partial, *a, **k):  # type: ignore
        try:
            if inspect.iscoroutinefunction(f):
                return await f(*a, **k)
            elif hasattr(f, "func") and inspect.iscoroutinefunction(f.func):
                return await f(*a, **k)
            else:
                return f(*a, **k)
        except XMPPError:
            raise
        except Exception as e:
            log.debug("Exception in %s", f, exc_info=e)
            raise XMPPError("internal-server-error", text=str(e))

    async def __wrap_form_handler(
        self,
        session: Optional["BaseSession[Any, Any]"],
        result: Form,
        form: SlixForm,
        adhoc_session: AdhocSessionType,
    ) -> AdhocSessionType:
        timer = self.__timeouts.pop(adhoc_session["id"], None)
        if timer is not None:
            print("canceled", adhoc_session["id"])
            timer.cancel()
        form_values = result.get_values(form)
        new_result = await self.__wrap_handler(
            result.handler,
            form_values,
            session,
            adhoc_session["from"],
            *result.handler_args,
            **result.handler_kwargs,
        )

        return await self.__handle_result(session, new_result, adhoc_session)

    async def __wrap_confirmation(
        self,
        session: Optional["BaseSession[Any, Any]"],
        confirmation: Confirmation,
        form: SlixForm,
        adhoc_session: AdhocSessionType,
    ) -> AdhocSessionType:
        if form.get_values().get("confirm"):  # type: ignore[no-untyped-call]
            result = await self.__wrap_handler(
                confirmation.handler,
                session,
                adhoc_session["from"],
                *confirmation.handler_args,
                **confirmation.handler_kwargs,
            )
            if confirmation.success:
                result = confirmation.success
        else:
            result = "You canceled the operation"

        return await self.__handle_result(session, result, adhoc_session)

    def register(self, command: Command, jid: JID | None = None) -> None:
        """
        Register a command as a adhoc command.

        this does not need to be called manually, ``BaseGateway`` takes care of
        that.

        :param command:
        :param jid:
        """
        if jid is None:
            jid = self.xmpp.boundjid
        elif not isinstance(jid, JID):
            jid = JID(jid)

        if (category := command.CATEGORY) is None:
            if command.NODE in self._commands:
                raise RuntimeError(
                    "There is already a command for the node '%s'", command.NODE
                )
            self._commands[command.NODE] = command
            self.xmpp.plugin["xep_0050"].add_command(  # type: ignore[no-untyped-call]
                jid=jid,
                node=command.NODE,
                name=strip_leading_emoji_if_needed(command.NAME),
                handler=partial(self.__wrap_initial_handler, command),
            )
        else:
            if isinstance(category, str):
                category = CommandCategory(category, category)
            node = category.node
            name = category.name
            if node not in self._categories:
                self._categories[node] = list[Command]()
                self.xmpp.plugin["xep_0050"].add_command(  # type: ignore[no-untyped-call]
                    jid=jid,
                    node=node,
                    name=strip_leading_emoji_if_needed(name),
                    handler=partial(self.__handle_category_list, category),
                )
            self._categories[node].append(command)

    async def get_items(self, jid: JID, node: str, iq: Iq) -> DiscoItems:
        """
        Get items for a disco query

        :param jid: the entity that should return its items
        :param node: which command node is requested
        :param iq:  the disco query IQ
        :return: commands accessible to the given JID will be listed
        """
        ifrom = iq.get_from()
        ifrom_str = str(ifrom)
        if (
            not self.xmpp.jid_validator.match(ifrom_str)
            and ifrom_str not in config.ADMINS
        ):
            raise XMPPError(
                "forbidden",
                "You are not authorized to execute adhoc commands on this gateway. "
                "If this is unexpected, ask your administrator to verify that "
                "'user-jid-validator' is correctly set in slidge's configuration.",
            )

        all_items = self.xmpp.plugin["xep_0030"].static.get_items(jid, node, None, None)
        log.debug("Static items: %r", all_items)
        if not all_items:
            return DiscoItems()

        session = self.xmpp.get_session_from_jid(ifrom)

        filtered_items = DiscoItems()
        filtered_items["node"] = self.xmpp.plugin["xep_0050"].stanza.Command.namespace
        for item in all_items:
            authorized = True
            if item["node"] in self._categories:
                for command in self._categories[item["node"]]:
                    try:
                        command.raise_if_not_authorized(
                            ifrom, fetch_session=False, session=session
                        )
                    except XMPPError:
                        authorized = False
                    else:
                        authorized = True
                        break
            else:
                try:
                    self._commands[item["node"]].raise_if_not_authorized(
                        ifrom, fetch_session=False, session=session
                    )
                except XMPPError:
                    authorized = False

            if authorized:
                filtered_items.append(item)

        return filtered_items


def strip_leading_emoji_if_needed(text: str) -> str:
    if config.STRIP_LEADING_EMOJI_ADHOC:
        return strip_leading_emoji(text)
    return text


log = logging.getLogger(__name__)



"""
The gateway
"""

from slixmpp import JID

from slidge import BaseGateway, FormField, GatewayUser
from slidge.command.register import RegistrationType

from .legacy_client import SuperDuperClient
from .util import ASSETS_DIR


class Gateway(BaseGateway):
    """
    This is instantiated once by the slidge entrypoint.

    By customizing the class attributes, we customize the registration process,
    and display name of the component.
    """

    COMPONENT_NAME = "The great legacy network (slidge)"
    COMPONENT_AVATAR = ASSETS_DIR / "slidge-color.png"
    COMPONENT_TYPE = "whatsapp"
    REGISTRATION_INSTRUCTIONS = (
        "Register to this fake service by using 'slidger' as username, and any "
        "password you want. Then you will need to enter '666' as the 2FA code."
    )
    REGISTRATION_TYPE = RegistrationType.TWO_FACTOR_CODE
    REGISTRATION_FIELDS = [
        FormField(var="username", label="User name", required=True),
        FormField(var="password", label="Password", required=True, private=True),
    ]
    GROUPS = True
    MARK_ALL_MESSAGES = True

    LEGACY_CONTACT_ID_TYPE = int

    async def validate(self, user_jid: JID, registration_form: dict[str, str | None]):
        """
        This function receives the values of the form defined in
        :attr:`REGISTRATION_FIELDS`. Here, since we set
        :attr:`REGISTRATION_TYPE` to "2FA", if this method does not raise any
        exception, the wannabe user will be prompted for their 2FA code.

        :param user_jid:
        :param registration_form:
        :return:
        """
        await SuperDuperClient.send_2fa(
            registration_form["username"],
            registration_form["password"],
        )

    async def validate_two_factor_code(self, user: GatewayUser, code: str):
        """
        This function receives the 2FA code entered by the aspiring user.

        It should raise something if the 2FA does not permit logging in to the
        legacy service.

        :param user:
        :param code:
        """
        await SuperDuperClient.validate_2fa(
            user.legacy_module_data["username"],
            user.legacy_module_data["password"],
            code,
        )



import asyncio
from pathlib import Path

ASSETS_DIR = Path(__file__).parent / "assets"
if not ASSETS_DIR.exists():
    ASSETS_DIR = Path(__file__).parent.parent / "dev" / "assets"
    if not ASSETS_DIR.exists():
        raise FileNotFoundError


def later(awaitable):
    asyncio.create_task(_later(awaitable))


async def _later(awaitable):
    await asyncio.sleep(0.5)
    await awaitable



"""
User actions
"""

from collections.abc import Iterable
from typing import TYPE_CHECKING, Union

from slidge import BaseSession, GatewayUser
from slidge.util.types import LinkPreview, Mention

from .group import MUC, Participant
from .legacy_client import SuperDuperClient

if TYPE_CHECKING:
    from .contact import Contact

Recipient = Union["Contact", MUC]
Sender = Union["Contact", Participant]


class Session(BaseSession[str, Recipient]):
    def __init__(self, user: GatewayUser):
        super().__init__(user)
        self.legacy_client = SuperDuperClient(self)
        self.contacts.user_legacy_id = self.legacy_client.user_id

    async def login(self):
        await self.legacy_client.login()
        return "Success!"

    async def on_text(
        self,
        chat: Recipient,
        text: str,
        *,
        reply_to_msg_id: str | None = None,
        reply_to_fallback_text: str | None = None,
        reply_to: Sender | None = None,  # type:ignore
        thread: str | None = None,  # type:ignore
        link_previews: Iterable[LinkPreview] = (),
        mentions: list[Mention] | None = None,
    ) -> str | None:
        if chat.is_group:
            assert isinstance(chat, MUC)
            msg = await self.legacy_client.send_group_msg(text, chat.legacy_id)
        else:
            msg = await self.legacy_client.send_direct_msg(text, chat.legacy_id)
        return msg.id



import logging

from slixmpp import (
    JID,
    CoroutineCallback,
    Iq,
    MatchXMLMask,
    Message,
    Presence,
    StanzaPath,
)
from slixmpp.exceptions import XMPPError

from ..util import DispatcherMixin, exceptions_to_xmpp_errors


class MucMiscMixin(DispatcherMixin):
    __slots__: list[str] = []

    def __init__(self, xmpp) -> None:
        super().__init__(xmpp)
        xmpp.register_handler(
            CoroutineCallback(
                "ibr_remove", StanzaPath("/iq/register"), self.on_ibr_remove
            )
        )

        xmpp.add_event_handler("groupchat_join", self.on_groupchat_join)
        xmpp.add_event_handler(
            "groupchat_direct_invite", self.on_groupchat_direct_invite
        )
        xmpp.add_event_handler("groupchat_subject", self.on_groupchat_subject)
        xmpp.add_event_handler("groupchat_message_error", self.__on_group_chat_error)
        xmpp.register_handler(
            CoroutineCallback(
                "muc_thread_subject",
                MatchXMLMask(
                    "<message xmlns='jabber:component:accept' type='groupchat'><subject/><thread/></message>"
                ),
                self.on_thread_subject,
            )
        )

    async def __on_group_chat_error(self, msg: Message) -> None:
        condition = msg["error"].get_condition()
        if condition not in KICKABLE_ERRORS:
            return

        try:
            muc = await self.get_muc_from_stanza(msg)
        except XMPPError as e:
            log.debug("Not removing resource", exc_info=e)
            return
        mfrom = msg.get_from()
        resource = mfrom.resource
        try:
            muc.remove_user_resource(resource)
        except KeyError:
            # this actually happens quite frequently on for both beagle and monal
            # (not sure why?), but is of no consequence
            log.debug("%s was not in the resources of %s", resource, muc)
        else:
            log.debug(
                "Removed %s from the resources of %s because of error", resource, muc
            )

    @exceptions_to_xmpp_errors
    async def on_ibr_remove(self, iq: Iq):
        if iq.get_to() == self.xmpp.boundjid.bare:
            return

        if iq["type"] == "set" and iq["register"]["remove"]:
            muc = await self.get_muc_from_stanza(iq)
            await muc.session.on_leave_group(muc.legacy_id)
            iq.reply().send()
            await muc.session.bookmarks.remove(
                muc, "You left this chat from an XMPP client."
            )
            return

        raise XMPPError("feature-not-implemented")

    @exceptions_to_xmpp_errors
    async def on_groupchat_join(self, p: Presence):
        if not self.xmpp.GROUPS:
            raise XMPPError(
                "feature-not-implemented",
                "This gateway does not implement multi-user chats.",
            )
        muc = await self.get_muc_from_stanza(p)
        await muc.join(p)

    @exceptions_to_xmpp_errors
    async def on_groupchat_direct_invite(self, msg: Message):
        invite = msg["groupchat_invite"]
        jid = JID(invite["jid"])

        if jid.domain != self.xmpp.boundjid.bare:
            raise XMPPError(
                "bad-request",
                "Legacy contacts can only be invited to legacy groups, not standard XMPP MUCs.",
            )

        if invite["password"]:
            raise XMPPError(
                "bad-request", "Password-protected groups are not supported"
            )

        session = await self._get_session(msg, logged=True)
        contact = await session.contacts.by_jid(msg.get_to())
        muc = await session.bookmarks.by_jid(jid)

        await session.on_invitation(contact, muc, invite["reason"] or None)

    @exceptions_to_xmpp_errors
    async def on_groupchat_subject(self, msg: Message):
        muc = await self.get_muc_from_stanza(msg)
        if not muc.HAS_SUBJECT:
            raise XMPPError(
                "bad-request",
                "There are no room subject in here. "
                "Use the room configuration to update its name or description",
            )
        await muc.on_set_subject(msg["subject"])

    @exceptions_to_xmpp_errors
    async def on_thread_subject(self, msg: Message):
        if msg["body"]:
            return
        session, muc, thread = await self._get_session_recipient_thread(msg)
        assert thread is not None
        await muc.on_set_thread_subject(thread, msg["subject"])  # type:ignore[union-attr]


KICKABLE_ERRORS = {
    "gone",
    "internal-server-error",
    "item-not-found",
    "jid-malformed",
    "recipient-unavailable",
    "redirect",
    "remote-server-not-found",
    "remote-server-timeout",
    "service-unavailable",
    "malformed error",
}

log = logging.getLogger(__name__)



import warnings
from datetime import datetime
from enum import IntEnum

import sqlalchemy as sa
from slixmpp import JID
from slixmpp.types import MucAffiliation, MucRole
from sqlalchemy import JSON, ForeignKey, Index, UniqueConstraint
from sqlalchemy.orm import Mapped, mapped_column, relationship

from ..util.types import ClientType, Hat, MucType
from .meta import Base, JSONSerializable, JSONSerializableTypes


class ArchivedMessageSource(IntEnum):
    """
    Whether an archived message comes from ``LegacyMUC.backfill()`` or was received
    as a "live" message.
    """

    LIVE = 1
    BACKFILL = 2


class GatewayUser(Base):
    """
    A user, registered to the gateway component.
    """

    __tablename__ = "user_account"
    id: Mapped[int] = mapped_column(primary_key=True)
    jid: Mapped[JID] = mapped_column(unique=True)
    registration_date: Mapped[datetime] = mapped_column(
        sa.DateTime, server_default=sa.func.now()
    )

    legacy_module_data: Mapped[JSONSerializable] = mapped_column(default={})
    """
    Arbitrary non-relational data that legacy modules can use
    """
    preferences: Mapped[JSONSerializable] = mapped_column(default={})
    avatar_hash: Mapped[str | None] = mapped_column(default=None)
    """
    Hash of the user's avatar, to avoid re-publishing the same avatar on the
    legacy network
    """

    contacts: Mapped[list["Contact"]] = relationship(
        back_populates="user", cascade="all, delete-orphan"
    )
    rooms: Mapped[list["Room"]] = relationship(
        back_populates="user", cascade="all, delete-orphan"
    )
    attachments: Mapped[list["Attachment"]] = relationship(cascade="all, delete-orphan")

    def __repr__(self) -> str:
        return f"User(id={self.id!r}, jid={self.jid!r})"

    def get(self, field: str, default: str = "") -> JSONSerializableTypes:
        # """
        # Get fields from the registration form (required to comply with slixmpp backend protocol)
        #
        # :param field: Name of the field
        # :param default: Default value to return if the field is not present
        #
        # :return: Value of the field
        # """
        return self.legacy_module_data.get(field, default)

    @property
    def registration_form(self) -> dict:
        # Kept for retrocompat, should be
        # FIXME: delete me
        warnings.warn(
            "GatewayUser.registration_form is deprecated.", DeprecationWarning
        )
        return self.legacy_module_data


class Avatar(Base):
    """
    Avatars of contacts, rooms and participants.

    To comply with XEPs, we convert them all to PNG before storing them.
    """

    __tablename__ = "avatar"

    id: Mapped[int] = mapped_column(primary_key=True)

    hash: Mapped[str] = mapped_column(unique=True)
    height: Mapped[int] = mapped_column()
    width: Mapped[int] = mapped_column()

    legacy_id: Mapped[str | None] = mapped_column(unique=True, nullable=True)

    # this is only used when avatars are available as HTTP URLs and do not
    # have a legacy_id
    url: Mapped[str | None] = mapped_column(unique=True, default=None)
    etag: Mapped[str | None] = mapped_column(default=None)
    last_modified: Mapped[str | None] = mapped_column(default=None)

    contacts: Mapped[list["Contact"]] = relationship(back_populates="avatar")
    rooms: Mapped[list["Room"]] = relationship(back_populates="avatar")


class Contact(Base):
    """
    Legacy contacts
    """

    __tablename__ = "contact"
    __table_args__ = (
        UniqueConstraint(
            "user_account_id", "legacy_id", name="uq_contact_user_account_id_legacy_id"
        ),
        UniqueConstraint(
            "user_account_id", "jid", name="uq_contact_user_account_id_jid"
        ),
    )

    id: Mapped[int] = mapped_column(primary_key=True)
    user_account_id: Mapped[int] = mapped_column(ForeignKey("user_account.id"))
    user: Mapped[GatewayUser] = relationship(lazy=True, back_populates="contacts")
    legacy_id: Mapped[str] = mapped_column(nullable=False)

    jid: Mapped[JID] = mapped_column()

    avatar_id: Mapped[int | None] = mapped_column(
        ForeignKey("avatar.id"), nullable=True
    )
    avatar: Mapped[Avatar | None] = relationship(lazy=False, back_populates="contacts")

    nick: Mapped[str | None] = mapped_column(nullable=True)

    cached_presence: Mapped[bool] = mapped_column(default=False)
    last_seen: Mapped[datetime | None] = mapped_column(nullable=True)
    ptype: Mapped[str | None] = mapped_column(nullable=True)
    pstatus: Mapped[str | None] = mapped_column(nullable=True)
    pshow: Mapped[str | None] = mapped_column(nullable=True)
    caps_ver: Mapped[str | None] = mapped_column(nullable=True)

    is_friend: Mapped[bool] = mapped_column(default=False)
    added_to_roster: Mapped[bool] = mapped_column(default=False)
    sent_order: Mapped[list["ContactSent"]] = relationship(
        back_populates="contact", cascade="all, delete-orphan"
    )

    extra_attributes: Mapped[JSONSerializable | None] = mapped_column(
        default=None, nullable=True
    )
    updated: Mapped[bool] = mapped_column(default=False)

    vcard: Mapped[str | None] = mapped_column()
    vcard_fetched: Mapped[bool] = mapped_column(default=False)

    participants: Mapped[list["Participant"]] = relationship(back_populates="contact")

    client_type: Mapped[ClientType] = mapped_column(nullable=False, default="pc")

    messages: Mapped[list["DirectMessages"]] = relationship(
        cascade="all, delete-orphan"
    )
    threads: Mapped[list["DirectThreads"]] = relationship(cascade="all, delete-orphan")


class ContactSent(Base):
    """
    Keep track of XMPP msg ids sent by a specific contact for networks in which
    all messages need to be marked as read.

    (XMPP displayed markers convey a "read up to here" semantic.)
    """

    __tablename__ = "contact_sent"
    __table_args__ = (
        UniqueConstraint(
            "contact_id", "msg_id", name="uq_contact_sent_contact_id_msg_id"
        ),
    )

    id: Mapped[int] = mapped_column(primary_key=True)
    contact_id: Mapped[int] = mapped_column(ForeignKey("contact.id"))
    contact: Mapped[Contact] = relationship(back_populates="sent_order")
    msg_id: Mapped[str] = mapped_column()


class Room(Base):
    """
    Legacy room
    """

    __table_args__ = (
        UniqueConstraint(
            "user_account_id", "legacy_id", name="uq_room_user_account_id_legacy_id"
        ),
        UniqueConstraint("user_account_id", "jid", name="uq_room_user_account_id_jid"),
    )

    __tablename__ = "room"
    id: Mapped[int] = mapped_column(primary_key=True)
    user_account_id: Mapped[int] = mapped_column(ForeignKey("user_account.id"))
    user: Mapped[GatewayUser] = relationship(lazy=True, back_populates="rooms")
    legacy_id: Mapped[str] = mapped_column(nullable=False)

    jid: Mapped[JID] = mapped_column(nullable=False)

    avatar_id: Mapped[int | None] = mapped_column(
        ForeignKey("avatar.id"), nullable=True
    )
    avatar: Mapped[Avatar | None] = relationship(lazy=False, back_populates="rooms")

    name: Mapped[str | None] = mapped_column(nullable=True)
    description: Mapped[str | None] = mapped_column(nullable=True)
    subject: Mapped[str | None] = mapped_column(nullable=True)
    subject_date: Mapped[datetime | None] = mapped_column(nullable=True)
    subject_setter: Mapped[str | None] = mapped_column(nullable=True)

    n_participants: Mapped[int | None] = mapped_column(default=None)

    muc_type: Mapped[MucType] = mapped_column(default=MucType.CHANNEL)

    user_nick: Mapped[str | None] = mapped_column()
    user_resources: Mapped[str | None] = mapped_column(nullable=True)

    participants_filled: Mapped[bool] = mapped_column(default=False)
    history_filled: Mapped[bool] = mapped_column(default=False)

    extra_attributes: Mapped[JSONSerializable | None] = mapped_column(default=None)
    updated: Mapped[bool] = mapped_column(default=False)

    participants: Mapped[list["Participant"]] = relationship(
        back_populates="room",
        primaryjoin="Participant.room_id == Room.id",
        cascade="all, delete-orphan",
    )

    archive: Mapped[list["ArchivedMessage"]] = relationship(
        cascade="all, delete-orphan"
    )

    messages: Mapped[list["GroupMessages"]] = relationship(cascade="all, delete-orphan")
    threads: Mapped[list["GroupThreads"]] = relationship(cascade="all, delete-orphan")


class ArchivedMessage(Base):
    """
    Messages of rooms, that we store to act as a MAM server
    """

    __tablename__ = "mam"
    __table_args__ = (
        UniqueConstraint("room_id", "stanza_id", name="uq_mam_room_id_stanza_id"),
    )

    id: Mapped[int] = mapped_column(primary_key=True)
    room_id: Mapped[int] = mapped_column(ForeignKey("room.id"), nullable=False)
    room: Mapped[Room] = relationship(lazy=True, back_populates="archive")

    stanza_id: Mapped[str] = mapped_column(nullable=False)
    timestamp: Mapped[datetime] = mapped_column(nullable=False)
    author_jid: Mapped[JID] = mapped_column(nullable=False)
    source: Mapped[ArchivedMessageSource] = mapped_column(nullable=False)
    legacy_id: Mapped[str | None] = mapped_column(nullable=True)

    stanza: Mapped[str] = mapped_column(nullable=False)

    displayed_by_user: Mapped[bool] = mapped_column(default=False, nullable=True)


class _LegacyToXmppIdsBase:
    """
    XMPP-client generated IDs, and mapping to the corresponding legacy IDs.

    A single legacy ID can map to several XMPP ids.
    """

    id: Mapped[int] = mapped_column(primary_key=True)
    legacy_id: Mapped[str] = mapped_column(nullable=False)
    xmpp_id: Mapped[str] = mapped_column(nullable=False)


class DirectMessages(_LegacyToXmppIdsBase, Base):
    __tablename__ = "direct_msg"
    __table_args__ = (Index("ix_direct_msg_legacy_id", "legacy_id", "foreign_key"),)
    foreign_key: Mapped[int] = mapped_column(ForeignKey("contact.id"), nullable=False)


class GroupMessages(_LegacyToXmppIdsBase, Base):
    __tablename__ = "group_msg"
    __table_args__ = (Index("ix_group_msg_legacy_id", "legacy_id", "foreign_key"),)
    foreign_key: Mapped[int] = mapped_column(ForeignKey("room.id"), nullable=False)


class GroupMessagesOrigin(_LegacyToXmppIdsBase, Base):
    """
    This maps "origin ids" <message id=XXX> to legacy message IDs
    We need that for message corrections and retractions, which do not reference
    messages by their "Unique and Stable Stanza IDs (XEP-0359)"
    """

    __tablename__ = "group_msg_origin"
    __table_args__ = (
        Index("ix_group_msg_origin_legacy_id", "legacy_id", "foreign_key"),
    )
    foreign_key: Mapped[int] = mapped_column(ForeignKey("room.id"), nullable=False)


class DirectThreads(_LegacyToXmppIdsBase, Base):
    __tablename__ = "direct_thread"
    __table_args__ = (Index("ix_direct_direct_thread_id", "legacy_id", "foreign_key"),)
    foreign_key: Mapped[int] = mapped_column(ForeignKey("contact.id"), nullable=False)


class GroupThreads(_LegacyToXmppIdsBase, Base):
    __tablename__ = "group_thread"
    __table_args__ = (Index("ix_direct_group_thread_id", "legacy_id", "foreign_key"),)
    foreign_key: Mapped[int] = mapped_column(ForeignKey("room.id"), nullable=False)


class Attachment(Base):
    """
    Legacy attachments
    """

    __tablename__ = "attachment"
    __table_args__ = (
        UniqueConstraint(
            "user_account_id",
            "legacy_file_id",
            name="uq_attachment_user_account_id_legacy_file_id",
        ),
    )

    id: Mapped[int] = mapped_column(primary_key=True)
    user_account_id: Mapped[int] = mapped_column(ForeignKey("user_account.id"))
    user: Mapped[GatewayUser] = relationship(back_populates="attachments")

    legacy_file_id: Mapped[str | None] = mapped_column(index=True, nullable=True)
    url: Mapped[str] = mapped_column(index=True, nullable=False)
    sims: Mapped[str | None] = mapped_column()
    sfs: Mapped[str | None] = mapped_column()


class Participant(Base):
    __tablename__ = "participant"
    __table_args__ = (
        UniqueConstraint("room_id", "resource", name="uq_participant_room_id_resource"),
        UniqueConstraint(
            "room_id", "contact_id", name="uq_participant_room_id_contact_id"
        ),
        UniqueConstraint(
            "room_id", "occupant_id", name="uq_participant_room_id_occupant_id"
        ),
    )

    id: Mapped[int] = mapped_column(primary_key=True)

    room_id: Mapped[int] = mapped_column(ForeignKey("room.id"), nullable=False)
    room: Mapped[Room] = relationship(
        lazy=False, back_populates="participants", primaryjoin=Room.id == room_id
    )

    contact_id: Mapped[int | None] = mapped_column(
        ForeignKey("contact.id"), nullable=True
    )
    contact: Mapped[Contact | None] = relationship(
        lazy=False, back_populates="participants"
    )

    occupant_id: Mapped[str] = mapped_column(nullable=False)

    is_user: Mapped[bool] = mapped_column(default=False)

    affiliation: Mapped[MucAffiliation] = mapped_column(
        default="member", nullable=False
    )
    role: Mapped[MucRole] = mapped_column(default="participant", nullable=False)

    presence_sent: Mapped[bool] = mapped_column(default=False)

    resource: Mapped[str] = mapped_column(nullable=False)
    nickname: Mapped[str] = mapped_column(nullable=False, default=None)
    nickname_no_illegal: Mapped[str] = mapped_column(nullable=False, default=None)

    hats: Mapped[list[Hat]] = mapped_column(JSON, default=list)

    extra_attributes: Mapped[JSONSerializable | None] = mapped_column(default=None)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.role = "participant"
        self.affiliation = "member"


class Bob(Base):
    __tablename__ = "bob"

    id: Mapped[int] = mapped_column(primary_key=True)
    file_name: Mapped[str] = mapped_column(nullable=False)

    sha_1: Mapped[str] = mapped_column(nullable=False, unique=True)
    sha_256: Mapped[str] = mapped_column(nullable=False, unique=True)
    sha_512: Mapped[str] = mapped_column(nullable=False, unique=True)

    content_type: Mapped[str | None] = mapped_column(nullable=False)
